from nose.tools import eq_, istest, raises
from jasmin_cis.data_io.hdf import __read_hdf4
from jasmin_cis.exceptions import InvalidVariableError
from jasmin_cis.test.test_files.data import *

@istest
@raises(IOError)
def should_raise_io_error_with_non_hdf_file():
    __read_hdf4(valid_cloud_cci_filename, valid_cloud_cci_variable)

@istest
def test_read_hdf4():
    filename = valid_hdf_sd_file
    sds, vds = __read_hdf4(filename,['Solution_Ocean','Path_Radiance_Land','Mean_Reflectance_Land'])

    # VD variable are listed in the VD part of the tuple, but not in the SD part
    eq_(True,vds.has_key('Solution_Ocean'))
    eq_(False,sds.has_key('Solution_Ocean'))

    # SD variable are listed in the SD part of the tuple, but not in the VD part
    eq_(True,sds.has_key('Path_Radiance_Land'))
    eq_(False,vds.has_key('Path_Radiance_Land'))
    eq_(True,sds.has_key('Mean_Reflectance_Land'))
    eq_(False,vds.has_key('Mean_Reflectance_Land'))

@istest
@raises(InvalidVariableError)
def test_that_cannot_read_unknown_variables():
    filename = valid_hdf_sd_file
    sds, vds = __read_hdf4(filename,['athing','unechose','einding'])

@istest
@raises(InvalidVariableError)
def test_that_cannot_read_unknown_variables_and_valid_variables():
    filename = valid_hdf_sd_file
    sds, vds = __read_hdf4(filename,['someBizarreVariableNobodyKnowsAbout','Solution_Ocean','Path_Radiance_Land','Mean_Reflectance_Land'])



'''
module to test the abstract AProduct class and it's helper methods
'''
from nose.tools import istest, eq_, raises, nottest
from jasmin_cis.test.test_files.data import *
from jasmin_cis.data_io.products.AProduct import get_data, __get_class
from jasmin_cis.exceptions import ClassNotFoundError

@istest
def test_that_get_data_accepts_valid_product():
    __get_class(valid_cloudsat_RVOD_file, product='Cloudsat_2B_CWC_RVOD')

@istest
def test_automatic_detection_of_product_for_existing_product():
    product_cls = __get_class(valid_cloudsat_RVOD_file)
    eq_(product_cls.__name__,'Cloudsat_2B_CWC_RVOD')

    product_cls = __get_class(valid_1d_filename)
    eq_(product_cls.__name__,'Xglnwa_vprof')

@istest
@raises(ClassNotFoundError)
def test_that_get_class_raises_ClassNotFoundError_for_non_existing_product():
    product_cls = __get_class('some_file_that_does_not_match_anything')

@istest
@raises(ClassNotFoundError)
def test_that_get_data_raises_ClassNotFoundError_for_missing_product():
    get_data(valid_cloudsat_RVOD_file,[valid_cloudsat_RVOD_variable], product='Product_Not_Yet_Implemented')

@istest
@raises(TypeError)
def test_that_get_data_throws_TypeError_for_invalid_product():
    from jasmin_cis.data_io.products.AProduct import AProduct
    # Create a class which subclasses AProduct but doesn't fully implement the interface
    class My_test_product(AProduct): pass
    my_product = My_test_product()



'''
 Module to test the colocation routines
'''
from nose.tools import istest, eq_, assert_almost_equal, raises
from jasmin_cis.exceptions import ClassNotFoundError
from test_util import mock


def is_colocated(data1, data2):
    '''
        Checks wether two datasets share all of the same points, this might be useful
        to determine if colocation is necesary or completed succesfully
    '''
    colocated = len(data1) == len(data2)
    if colocated:
        for i, point1 in enumerate(data1):
            colocated = point1.same_point_in_space_and_time(data2[i])
            if not colocated:
                return colocated
    return colocated


class ColocatorTests(object):
    pass


class TestDefaultColocator(ColocatorTests):
    pass


class TestAverageColocator(ColocatorTests):

    @istest
    @raises(ClassNotFoundError)
    def should_throw_error_if_full_average_kernel_is_not_used(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import AverageColocator, mean, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34)))

        col = AverageColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), mean())


    @istest
    def test_basic_col_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import AverageColocator, full_average, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34)))

        col = AverageColocator(stddev_name='my_std_dev',nopoints_name='my_no_points')
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), full_average())
        means = new_data[0]
        std_dev = new_data[1]
        no_points = new_data[2]

        eq_(means.name(), 'rain_mean')
        eq_(std_dev.name(), 'my_std_dev')
        eq_(no_points.name(), 'my_no_points')


class TestDifferenceColocator(ColocatorTests):

    @istest
    def test_basic_col_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DifferenceColocator, mean, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34),25.0))

        col = DifferenceColocator(diff_name='my_diff')
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), mean())
        values = new_data[0]
        differences = new_data[1]

        eq_(values.name(), 'rain')
        eq_(differences.name(), 'my_diff')
        eq_(values.data[0],25.5)
        eq_(differences.data[0],0.5)



class KernelTests(object):
    pass


class Test_full_average(KernelTests):

    @istest
    def test_basic_col_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import AverageColocator, full_average, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34)))

        col = AverageColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), full_average())
        means = new_data[0]
        std_dev = new_data[1]
        no_points = new_data[2]

        eq_(means.data[0], 25.5)
        assert_almost_equal(std_dev.data[0], 14.4308696897)
        eq_(no_points.data[0], 50)


class Test_nn_gridded(KernelTests):

    @istest
    def test_basic_col_gridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_gridded, DummyConstraint
        cube = mock.make_square_5x3_2d_cube()
        sample_points = [ HyperPoint(1.0, 1.0), HyperPoint(4.0,4.0), HyperPoint(-4.0,-4.0) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), nn_gridded())[0]
        eq_(new_data.data[0], 8.0)
        eq_(new_data.data[1], 12.0)
        eq_(new_data.data[2], 4.0)

    @istest
    def test_already_colocated_in_col_gridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_gridded, DummyConstraint
        cube = mock.make_square_5x3_2d_cube()
        # This point already exists on the cube with value 5 - which shouldn't be a problem
        sample_points = [ HyperPoint(0.0, 0.0) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), nn_gridded())[0]
        eq_(new_data.data[0], 8.0)

    @istest
    def test_coordinates_exactly_between_points_in_col_gridded_to_ungridded_in_2d(self):
        '''
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                Iris seems to count a point as 'belonging' to a datapoint if it is greater than a datapoint cell's lower
                bound and less than or equal to it's upper bound. Where a cell is an imaginary boundary around a datapoint
                which divides the grid.
        '''
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_gridded, DummyConstraint
        cube = mock.make_square_5x3_2d_cube()
        sample_points = [ HyperPoint(2.5, 2.5), HyperPoint(-2.5, 2.5), HyperPoint(2.5, -2.5), HyperPoint(-2.5, -2.5) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), nn_gridded())[0]
        eq_(new_data.data[0], 8.0)
        eq_(new_data.data[1], 5.0)
        eq_(new_data.data[2], 7.0)
        eq_(new_data.data[3], 4.0)

    @istest
    def test_coordinates_outside_grid_in_col_gridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_gridded, DummyConstraint
        cube = mock.make_square_5x3_2d_cube()
        sample_points = [ HyperPoint(5.5, 5.5), HyperPoint(-5.5, 5.5), HyperPoint(5.5, -5.5), HyperPoint(-5.5, -5.5) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), nn_gridded())[0]
        eq_(new_data.data[0], 12.0)
        eq_(new_data.data[1], 6.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)

    @istest
    def test_basic_col_gridded_to_ungridded_in_2d_with_time(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_gridded, DummyConstraint
        import datetime as dt
        cube = mock.make_square_5x3_2d_cube_with_time()

        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,t=dt.datetime(1984,8,28,8,34)))
        sample_points.append(HyperPoint(4.0,4.0,t=dt.datetime(1984,8,31,1,23)))
        sample_points.append(HyperPoint(-4.0,-4.0,t=dt.datetime(1984,9,2,15,54)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), nn_gridded())[0]
        eq_(new_data.data[0], 51.0)
        eq_(new_data.data[1], 82.0)
        eq_(new_data.data[2], 28.0)


class Test_nn_horizontal(KernelTests):

    @istest
    def test_basic_col_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_horizontal, DummyConstraint
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = [HyperPoint(1.0, 1.0), HyperPoint(4.0,4.0), HyperPoint(-4.0,-4.0)]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_horizontal())[0]
        eq_(new_data.data[0], 8.0)
        eq_(new_data.data[1], 12.0)
        eq_(new_data.data[2], 4.0)

    @istest
    def test_already_colocated_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_horizontal, DummyConstraint
        ug_data = mock.make_regular_2d_ungridded_data()
        # This point already exists on the cube with value 5 - which shouldn't be a problem
        sample_points = [ HyperPoint(0.0, 0.0) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_horizontal())[0]
        eq_(new_data.data[0], 8.0)

    @istest
    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        '''
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_horizontal algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly the same
                you would expect the first point to be chosen. This doesn't seem to always be the case but is probably
                down to floating points errors in the haversine calculation as these test points are pretty close
                together. This test is only really for documenting the behaviour for equidistant points.
        '''
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_horizontal, DummyConstraint
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = [ HyperPoint(2.5, 2.5), HyperPoint(-2.5, 2.5), HyperPoint(2.5, -2.5), HyperPoint(-2.5, -2.5) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_horizontal())[0]
        eq_(new_data.data[0], 11.0)
        eq_(new_data.data[1], 5.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)

    @istest
    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_horizontal, DummyConstraint
        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = [ HyperPoint(5.5, 5.5), HyperPoint(-5.5, 5.5), HyperPoint(5.5, -5.5), HyperPoint(-5.5, -5.5) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_horizontal())[0]
        eq_(new_data.data[0], 12.0)
        eq_(new_data.data[1], 6.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)

class Test_nn_time(KernelTests):

    @istest
    @raises(TypeError)
    def test_basic_col_with_incompatible_points_throws_a_TypeError(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        # Make sample points with no time dimension specified
        sample_points = [HyperPoint(1.0, 1.0), HyperPoint(4.0,4.0), HyperPoint(-4.0,-4.0)]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]

    @istest
    def test_basic_col_in_2d_with_time(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,t=dt.datetime(1984,8,29,8,34)))
        sample_points.append(HyperPoint(4.0,4.0,t=dt.datetime(1984,9,2,1,23)))
        sample_points.append(HyperPoint(-4.0,-4.0,t=dt.datetime(1984,9,4,15,54)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]
        eq_(new_data.data[0], 3.0)
        eq_(new_data.data[1], 7.0)
        eq_(new_data.data[2], 10.0)

    @istest
    def test_basic_col_with_time(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        import numpy as np

        ug_data = mock.make_MODIS_time_steps()

        ref = np.array([0.0,1.0,2.0,3.0])

        sample_points = HyperPointList()
        sample_points.append(HyperPoint(0.0, 0.0,t=149751.369618055))
        sample_points.append(HyperPoint(0.0,0.0,t=149759.378055556,))
        sample_points.append(HyperPoint(0.0,0.0,t=149766.373969907))
        sample_points.append(HyperPoint(0.0,0.0,t=149776.375995371))

        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]
        assert(np.equal(new_data.data,ref).all())

    @istest
    def test_already_colocated_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        import datetime as dt
        import numpy as np
        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()

        t0 = dt.datetime(1984,8,27)
        for d in xrange(15):
            sample_points.append(HyperPoint(0.0, 0.0,t=t0+dt.timedelta(days=d)))

        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]
        assert(np.equal(new_data.data,np.arange(15)+1.0).all())

    @istest
    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        '''
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_time algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same the first point to be chosen.
        '''
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()
        # Choose a time at midday
        sample_points.append(HyperPoint(0.0,0.0,t=dt.datetime(1984,8,29,12)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]
        eq_(new_data.data[0], 3.0)

    @istest
    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_time, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(0.0,0.0,t=dt.datetime(1984,8,26)))
        sample_points.append(HyperPoint(0.0,0.0,t=dt.datetime(1884,8,26)))
        sample_points.append(HyperPoint(0.0,0.0,t=dt.datetime(1994,8,27)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_time())[0]
        eq_(new_data.data[0], 1.0)
        eq_(new_data.data[1], 1.0)
        eq_(new_data.data[2], 15.0)


class Test_nn_altitude(KernelTests):

    @istest
    @raises(TypeError)
    def test_basic_col_with_incompatible_points_throws_a_TypeError(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import DefaultColocator, nn_vertical, DummyConstraint
        ug_data = mock.make_regular_4d_ungridded_data()
        # Make sample points with no time dimension specified
        sample_points = [HyperPoint(1.0, 1.0), HyperPoint(4.0,4.0), HyperPoint(-4.0,-4.0)]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_vertical())[0]

    @istest
    def test_basic_col_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_vertical, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34)))
        sample_points.append(HyperPoint(4.0,4.0,34.0,dt.datetime(1984,9,2,1,23)))
        sample_points.append(HyperPoint(-4.0,-4.0,89.0,dt.datetime(1984,9,4,15,54)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_vertical())[0]
        eq_(new_data.data[0], 6.0)
        eq_(new_data.data[1], 16.0)
        eq_(new_data.data[2], 46.0)

    @istest
    def test_already_colocated_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_vertical, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(0.0,0.0,80.0,dt.datetime(1984,9,4,15,54)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_vertical())[0]
        eq_(new_data.data[0], 41.0)

    @istest
    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        '''
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_time algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same the first point to be chosen.
        '''
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_vertical, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        # Choose a time at midday
        sample_points.append(HyperPoint(0.0,0.0,35.0,dt.datetime(1984,8,29,12)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_vertical())[0]
        eq_(new_data.data[0], 16.0)

    @istest
    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, nn_vertical, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(0.0, 0.0,-12.0,dt.datetime(1984,8,29,8,34)))
        sample_points.append(HyperPoint(0.0,0.0,91.0,dt.datetime(1984,9,2,1,23)))
        sample_points.append(HyperPoint(0.0,0.0,890.0,dt.datetime(1984,9,4,15,54)))
        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), nn_vertical())[0]
        eq_(new_data.data[0], 1.0)
        eq_(new_data.data[1], 46.0)
        eq_(new_data.data[2], 46.0)


class Test_mean(KernelTests):

    @istest
    def test_basic_col_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint, HyperPointList
        from jasmin_cis.col_implementations import DefaultColocator, mean, DummyConstraint
        import datetime as dt
        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(1.0, 1.0,12.0,dt.datetime(1984,8,29,8,34)))

        col = DefaultColocator()
        new_data = col.colocate(sample_points, ug_data, DummyConstraint(), mean())[0]
        eq_(new_data.data[0], 25.5)


class Test_li(KernelTests):

    @istest
    def test_basic_col_gridded_to_ungridded_using_li_in_2d(self):
        from jasmin_cis.col_implementations import DefaultColocator, li, DummyConstraint
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        cube = mock.make_square_5x3_2d_cube()
        sample_points = [ HyperPoint(1.0, 1.0), HyperPoint(4.0,4.0), HyperPoint(-4.0,-4.0) ]
        col = DefaultColocator()
        new_data = col.colocate(sample_points, cube, DummyConstraint(), li())[0]
        assert_almost_equal(new_data.data[0], 8.8)
        assert_almost_equal(new_data.data[1], 11.2)
        assert_almost_equal(new_data.data[2], 4.8)

class ConstraintTests(object):
    pass

class TestSepConstraint(ConstraintTests):

    @istest
    def test_all_constraint_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(0.0, 0.0, 50.0,dt.datetime(1984,8,29))

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        h_sep = 1000
        # 15m altitude seperation
        a_sep = 15
        # 1 day (and a little bit) time seperation
        t_sep = '1d1M'

        constraint = SepConstraint(h_sep = h_sep, a_sep = a_sep, t_sep = t_sep)

        # This should leave us with 9 points: [[ 22, 23, 24]
        #                                      [ 27, 28, 29]
        #                                      [ 32, 33, 34]]
        ref_vals = np.array([22,23,24,27,28,29,32,33,34])

        new_points = constraint.constrain_points(sample_point,ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert(np.equal(ref_vals, new_vals).all())

    @istest
    def test_alt_constraint_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(0.0, 0.0, 50.0,dt.datetime(1984,8,29))

        # 15m altitude seperation
        a_sep = 15

        constraint = SepConstraint(a_sep=a_sep)

        # This should leave us with 15 points:   [ 21.  22.  23.  24.  25.]
        #                                       [ 26.  27.  28.  29.  30.]
        #                                       [ 31.  32.  33.  34.  35.]
        ref_vals = np.array([21.,22.,23.,24.,25.,26.,27.,28.,29.,30.,31.,32.,33.,34.,35.])

        new_points = constraint.constrain_points(sample_point,ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert(np.equal(ref_vals, new_vals).all())

    @istest
    def test_horizontal_constraint_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(0.0, 0.0, 50.0,dt.datetime(1984,8,29))

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        #  in each direction
        constraint = SepConstraint(h_sep=1000)

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50)+1.0,(10,5))[:,1:4].flatten()

        new_points = constraint.constrain_points(sample_point,ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert(np.equal(ref_vals, new_vals).all())

    @istest
    def test_time_constraint_in_4d(self):
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        from jasmin_cis.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(0.0, 0.0, 50.0,dt.datetime(1984,8,29))

        # 1 day (and a little bit) time seperation
        constraint = SepConstraint(t_sep='1d1M')

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50)+1.0,(10,5))[:,1:4].flatten()

        new_points = constraint.constrain_points(sample_point,ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert(np.equal(ref_vals, new_vals).all())



__author__ = 'duncan'

from jasmin_cis.time_util import convert_sec_since_to_std_time, convert_std_time_to_datetime
import datetime as dt
from iris.unit import Unit

Caliop_time_unit = Unit('seconds since 1993-01-01', calendar='gregorian')

Caliop_UTC_time_unit = Unit('days since 2010-01-01', calendar='gregorian')

t1a = Caliop_time_unit.num2date(536539236.156000)

t1 = convert_sec_since_to_std_time(536539236.156000, dt.datetime(1993,1,1,0,0,0))

t2 = dt.datetime(2010,1,1) + dt.timedelta(days=0.944782)

t3 = Caliop_UTC_time_unit.num2date(0.944782)

print convert_std_time_to_datetime(t1)
print t1a
print t2
print t3


'''
module to test the various subclasses of the abstract AProduct class
'''
from nose.tools import istest, eq_, raises, nottest
from jasmin_cis.data_io.products.products import *
from jasmin_cis.exceptions import InvalidVariableError
from jasmin_cis.test.test_files.data import non_netcdf_file


def check_regex_matching(cls_name, filename):
    from jasmin_cis.data_io.products.AProduct import __get_class
    cls = __get_class(filename)
    eq_(cls.__name__,cls_name)

invalid_variable = "im_an_invalid_variable"
invalid_filename = "im_an_invalid_file"
invalid_format = non_netcdf_file

class ProductTests():

    @istest
    def test_file_regex_matching(self):
        check_regex_matching(self.product.__name__, self.filename)

    @istest
    def test_file_regex_matching_for_full_path(self):
        check_regex_matching(self.product.__name__, "/home/duncan/"+self.filename)
        check_regex_matching(self.product.__name__, "/a/xenida/more/lev20/confusing/hdf/path/nc/"+self.filename)

    @istest
    def test_create_data_object(self):
        self.product().create_data_object([self.filename], self.valid_variable)

    @istest
    def test_create_coords(self):
        valid_standard_names = ['latitude', 'longitude', 'altitude', 'time']
        coords = self.product().create_coords([self.filename])

        for coord in coords:
            logging.debug(coord.metadata.standard_name)

        assert(all([ coord.metadata.standard_name in valid_standard_names for coord in coords ]))

    @istest
    def test_write_coords(self):
        from jasmin_cis.data_io.write_netcdf import write_coordinates
        from os import remove
        test_file = '/tmp/test_out.nc'
        coords = self.product().create_coords([self.filename])
        write_coordinates(coords, test_file)
        remove(test_file)

    @istest
    @raises(IOError)
    def should_raise_ioerror_with_invalid_filename(self):
        self.product().create_data_object([invalid_filename], self.valid_variable)

    @istest
    @raises(IOError)
    def should_raise_ioerror_with_file_that_is_not_a_recognised_format(self):
        self.product().create_data_object([invalid_format], self.valid_variable)

    @istest
    @raises(InvalidVariableError)
    def should_raise_error_when_variable_does_not_exist_in_file(self):
        self.product().create_data_object([self.filename], invalid_variable)


class TestCloudsat(ProductTests):

    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_cloudsat_RVOD_file, valid_cloudsat_RVOD_variable
        self.filename = valid_cloudsat_RVOD_file
        self.valid_variable = valid_cloudsat_RVOD_variable
        self.product = Cloudsat_2B_CWC_RVOD

    @istest
    @raises(InvalidVariableError)
    def should_raise_error_when_variable_does_not_exist_in_file(self):
        # workaround for HDF library bug in Jasmin
        raise InvalidVariableError


class TestMODIS_L3(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_modis_l3_filename, valid_modis_l3_variable
        self.filename = valid_modis_l3_filename
        self.valid_variable = valid_modis_l3_variable
        self.product = MODIS_L3


class TestCaliop_L2(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_caliop_l2_filename, valid_caliop_l2_variable
        self.filename = valid_caliop_l2_filename
        self.valid_variable = valid_caliop_l2_variable
        self.product = Caliop_L2


class TestCaliop_L1(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_caliop_l1_filename, valid_caliop_l1_variable
        self.filename = valid_caliop_l1_filename
        self.valid_variable = valid_caliop_l1_variable
        self.product = Caliop_L1


class TestMODIS_L2(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_modis_l2_filename, valid_modis_l2_variable
        self.filename = valid_modis_l2_filename
        self.valid_variable = valid_modis_l2_variable
        self.product = MODIS_L2


class TestCloud_CCI(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_cloud_cci_filename, valid_cloud_cci_variable
        self.filename = valid_cloud_cci_filename
        self.valid_variable = valid_cloud_cci_variable
        self.product = Cloud_CCI

class TestAerosol_CCI(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_aerosol_cci_filename, valid_aerosol_cci_variable
        self.filename = valid_aerosol_cci_filename
        self.valid_variable = valid_aerosol_cci_variable
        self.product = Aerosol_CCI

class TestCis(ProductTests):

    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_cis_col_file
        self.filename = valid_cis_col_file
        self.valid_variable = 'AOT_440'
        self.product = Cis

class TestNCAR_NetCDF_RAF(ProductTests):

    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_NCAR_NetCDF_RAF_filename, valid_NCAR_NetCDF_RAF_variable
        self.filename = valid_NCAR_NetCDF_RAF_filename
        self.valid_variable = valid_NCAR_NetCDF_RAF_variable
        self.product = NCAR_NetCDF_RAF

# class TestXglnwa(ProductTests):
#
#     def __init__(self):
#         self.filename = valid_1d_filename
#         self.invalid_filename = "im_an_invalid_file"
#         self.invalid_format = non_netcdf_file
#         self.valid_variable = valid_variable_in_valid_filename
#         self.invalid_variable = "im_an_invalid_variable"
#         self.file_without_read_permissions = file_without_read_permissions
#         self.product = Xglnwa

class TestXglnwa_vprof(ProductTests):

    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_1d_filename, valid_variable_in_valid_filename
        self.filename = valid_1d_filename
        self.valid_variable = valid_variable_in_valid_filename
        self.product = Xglnwa_vprof


class TestXenida(ProductTests):

    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_xenida_filename, valid_xenida_variable
        self.filename = valid_xenida_filename
        self.valid_variable = valid_xenida_variable
        self.product = Xenida

    @nottest
    def test_write_coords(self):
        # This won't work for model data yet as the coordinates aren't all the same shape,
        #  they need to be 'expanded'
        pass

class TestAeronet(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_aeronet_filename, valid_aeronet_variable, another_valid_aeronet_filename
        self.filename = valid_aeronet_filename
        self.filenames = [valid_aeronet_filename, another_valid_aeronet_filename]
        self.valid_variable = valid_aeronet_variable
        self.product = Aeronet

    @istest
    def test_create_data_object_from_multiple_files(self):
        self.product().create_data_object(self.filenames, self.valid_variable)

class TestASCII(ProductTests):
    def __init__(self):
        from jasmin_cis.test.test_files.data import valid_ascii_filename, valid_ascii_variable, ascii_filename_with_no_values
        self.filename = valid_ascii_filename
        self.no_value_filename = ascii_filename_with_no_values
        self.valid_variable = valid_ascii_variable
        self.product = ASCII_Hyperpoints

    @istest
    @raises(IOError)
    def should_raise_error_when_variable_does_not_exist_in_file(self):
        '''
         This product throws an IO error rather than an InvalidVariable error as the file can only have one variable
        @return:
        '''
        self.product().create_data_object([self.no_value_filename], True)

    @istest
    def test_create_data_object_with_missing_values(self):
        '''
         Check that missing values get masked correctly
        @return:
        '''
        data = self.product().create_data_object([self.filename], True)
        assert(all(data.data.mask == [False, False, False, True, False, True, False, False]))

    @istest
    def test_create_data_object_with_valid_datetime(self):
        import datetime
        from jasmin_cis.time_util import convert_datetime_to_std_time
        data = self.product().create_data_object([self.filename], True)
        assert(data.coord('time').data[3] == convert_datetime_to_std_time(datetime.datetime(2012,8,25,15,32,03)))
        assert(data.coord('time').data[4] == convert_datetime_to_std_time(datetime.datetime(2012,8,26)))



from abc import ABCMeta, abstractmethod
import logging

class AProduct(object):
    '''
    Abstract class for the various possible data products. This just defines the interface which
    the subclasses must implement.
    '''
    __metaclass__ = ABCMeta

    @abstractmethod
    def create_data_object(self, filenames, variable):
        '''
        Create a an ungridded data object for a given variable from many files

        @param filenames:    List of filenames of files to read
        @param usr_variable:    Variable to read from the files
        @return: An UngriddedData object for the specified variable

        @raise FileIOError: Unable to read a file
        @raise InvalidVariableError: Variable not present in file
        '''

    @abstractmethod
    def create_coords(self, filenames):
        '''
        Reads the coordinates from a bunch of files
        @param filenames: List of filenames to read coordinates from
        @return: L{CoordList} object
        '''

    @abstractmethod
    def get_file_signature(self):
        '''
        @return: a list of regex to match the product's file naming convention.

        Example
        -------

        return [r'.*CODE*.nc']
        This will match all files with a name containing the string 'CODE' and with the 'nc' extension.
        '''


def __get_class(filename, product=None):
    '''
    Identify the subclass of L{AProduct} to a given product name if specified.
    If the product name is not specified, the routine uses the signature (regex)
    given by get_file_signature() to infer the product class from the filename.

    Note, only the first filename of the list is use here.

    @param filename: A single filename
    @param product: name of the product
    @return: a subclass of L{AProduct}
    '''
    import re
    import os
    import jasmin_cis.plugin as plugin
    from jasmin_cis.exceptions import ClassNotFoundError

    # Ensure the filename doesn't include the path
    filename = os.path.basename(filename)

    product_classes = plugin.find_plugin_classes(AProduct, 'jasmin_cis.data_io.products.products')

    for cls in product_classes:

        if product is None:
            # search for a pattern that matches file signature
            patterns = cls().get_file_signature()
            for pattern in patterns:
                # Match the pattern - re.I allows for case insensitive matches
                if re.match(pattern,filename,re.I) is not None:
                    logging.debug("Found product class " + cls.__name__ + " matching regex pattern " + pattern)
                    return cls
        else:
            # product specified directly
            if product == cls.__name__:
                logging.debug("Selected product class " +  cls.__name__)
                return cls
    error_message = "Product cannot be found for given file.\nSupported products and signatures are:\n"
    for cls in product_classes:
        error_message += cls().__class__.__name__ + ": " + str(cls().get_file_signature()) + "\n"
    raise ClassNotFoundError(error_message)


def get_data(filenames, variable, product=None):
    '''
    Top level routine for calling the correct product's create_ungridded_data routine.

    @param product: The product to read data from - this should be a string which matches the name of one of the subclasses of AProduct
    @param filenames: A list of filenames to read data from
    @param variable: The variable to create the UngriddedData object from
    @return: An Ungridded data variable
    '''
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving data using product " +  product_cls.__name__ + "...")
    data = product_cls().create_data_object(filenames, variable)
    return data


def get_coordinates(filenames, product=None):
    '''
    Top level routine for calling the correct product's create_coords routine.

    @param product: The product to read data from - this should be a string which matches the name of one of the subclasses of AProduct
    @param filenames: A list of filenames to read data from
    @return: A CoordList object
    '''
    product_cls = __get_class(filenames[0], product)

    logging.info("Retrieving coordinates using product " +  product_cls.__name__)
    coord = product_cls().create_coords(filenames)
    return coord


from pyhdf.SD import SD, SDC, SDAttr

def set_hdf4_attr(obj, name, value):
    import types
    from pyhdf.error import HDF4Error
    # Called by the __setattr__ method of the SD, SDS and SDim objects.

    # Be careful with private attributes.
    #if name in privAttr:
#    if name[0] == '_':
#        obj.__dict__[name] = value
#        return

    # Treat everything else as an HDF attribute.
    if type(value) not in [types.ListType, types.TupleType]:
        value = [value]
    typeList = []
    for v in value:
        t = type(v)
        # Prohibit mixing numeric types and strings.
        if t in [types.IntType, types.FloatType] and \
               not types.StringType in typeList:
            if t not in typeList:
                typeList.append(t)
        # Prohibit sequence of strings or a mix of numbers and string.
        elif t == types.StringType and not typeList:
            typeList.append(t)
        else:
            typeList = []
            break
    if types.StringType in typeList:
        xtype = SDC.CHAR8
        value = value[0]
    # double is "stronger" than int
    elif types.FloatType in typeList:
        xtype = SDC.FLOAT64
    elif types.IntType in typeList:
        xtype = SDC.INT32
    else:
        raise HDF4Error, "Illegal attribute value"

    # Assign value
    try:
        a = SDAttr(obj, name)
        a.set(xtype, value)
    except HDF4Error, msg:
        raise HDF4Error, "cannot set attribute: %s" % msg


def __create_variable(hdf_file, name, data, **kwargs):
    '''
    Create a variable in the given hdf file
    
    @param hdf_file:    The hdf file in which to create the variable
    @param name:        The name of the variable
    @param data:        An (numpy) array containing the data
    '''
    # Create the variable
    var = hdf_file.create(name, SDC.FLOAT32, len(data))
    
    # Set any extra attributes
    for attr_name, value in kwargs.items():
        #att = var.attr(attr_name)
        #att.set(type(value), value)
        set_hdf4_attr(var,attr_name, value)
    # Build up a tuple of the data points
    t = ()
    for point in data:
        t = t + (point,)
    # Assign the tuple of data points to the variable
    var[0] = t
    # End access to the variable
    var.endaccess()
    
def write(obj, filename):
    '''
    Writes an object to a file
    
    @param obj:        The ungridded data object to write
    @param filename:   The filename of the file to be written 
    '''
    # Create file, open in read-write mode, and overwrite any existing data
    hdf_file = SD(filename, SDC.WRITE|SDC.CREATE|SDC.TRUNC)
    
    # Create variables
    __create_variable(hdf_file, obj.standard_name, obj.data, _FillValue=obj.missing_value, units=str(obj.units), long_name='Model Data')
    __create_variable(hdf_file, obj.coords()[0].name().title(), obj.x)
    
    if obj.y is not None:
        __create_variable(hdf_file, obj.coords()[1].name().title(), obj.y)
    
    # Close file
    hdf_file.end()


from jasmin_cis.data_io.write_netcdf import *


def test_main():
    from jasmin_cis.data_io.ungridded_data import UngriddedData, Metadata
    from jasmin_cis.data_io.Coord import Coord
    from numpy import array
    coords = []
    #    coords.append(Coord(array([1,2,3,4,5,6,7,8,9,10]), Metadata(name="pixel_number",
    #                                                                long_name="Sinusoidal pixel index",
    #                                                                shape=(10,),
    #                                                                range = (0,90352))))
    coords.append(Coord(array([1,2,3,4,5,6,7,8,9,10]), Metadata(name="lon",
                                                                long_name="Longitude",
                                                                shape=(10,),
                                                                units="degrees_east",
                                                                range=(-180, 180),
                                                                missing_value=-999), "X"))
    coords.append(Coord(array([3,5,3,1,8,5,3,2,6,31]), Metadata(name="lat",
                                                                long_name="Latitude",
                                                                shape=(10,),
                                                                units="degrees_north",
                                                                range=(-90, 90),
                                                                missing_value=-999), "Y"))
    coords.append(Coord(array([7,8,9,10,11,12,13,14,15,16]), Metadata(name="time",
                                                                      long_name="TAI70 time",
                                                                      shape=(10,),
                                                                      units="seconds",
                                                                      range=(1,-1),
                                                                      missing_value=0), "T"))

    data = array([6,43,86,25,86,12,95,45,73,87])
    metadata = Metadata(name='mass_fraction_of_cloud_liquid_water_in_air',
                        long_name='Long Rain',
                        shape=(10,),
                        units='Rain units',
                        range=(0,100),
                        missing_value='-999')
    data_object = UngriddedData(data, metadata, coords)
    print data.dtype
    write(data_object, "ungridded_netcdf.nc")

def test_read():
    from jasmin_cis.data_io.products.AProduct import get_data

    filenames = ["ungridded_netcdf.nc"]
    data_object = get_data(filenames, 'mass_fraction_of_cloud_liquid_water_in_air',"Cloud_CCI")

    print data_object.data

    from plotting.plot import Plotter

    Plotter([data_object])


test_main()

test_read()


from abc import ABCMeta, abstractmethod


class Colocator(object):
    '''
    Class which provides a method for performing colocation. This just defines the interface which
    the subclasses must implement.
    '''
    __metaclass__ = ABCMeta

    @abstractmethod
    def colocate(self, points, data, constraint, kernel):
        '''

        @param points: A set of points onto which we will colocate some other 'data'
        @param data: Some other data to be colocated onto the 'points'
        @param constraint: A Constraint instance which provides a fill_value and constrain method
        @param kernel: A Kernel instance which provides a kernel method
        @return: One or more LazyData objects whose coordinates lie on the points defined above

        '''


class Kernel(object):
    '''
    Class which provides a method for taking a number of points and returning one value. This could be a nearest neighbour algorithm
     or some sort of algorithm. This just defines the interface which the subclasses must implement.
    '''
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_value(self, point, data):
        '''
        @param point: A single HyperPoint
        @param data: A set of data points to reduce to a single value
        @return: A single value (number) which represents some operation on the points provided

        '''


class Constraint(object):
    '''
    Class which provides a method for constraining a set of points. A single HyperPoint is given as a reference but the data points to be
     be reduced ultimately may be of any type. This just defines the interface which the subclasses must implement.
    '''
    __metaclass__ = ABCMeta

    def __init__(self):
        import numpy as np
        self.fill_value = np.Infinity

    @abstractmethod
    def constrain_points(self, point, data):
        '''

        @param point: A single HyperPoint
        @param data: A set of data points to be reduced
        @return: A reduced set of data points

        '''


def __get_class(parent_class, name=None):
    '''
    Identify the subclass of parent_class to a given name, if specified.
    If the name is not specified, the routine picks a default.

    Note, only the first filename of the list is use here.

    @param parent_class: A base class
    @param name: name of the class to find
    @return: a subclass of the parent_class
    '''
    from jasmin_cis.plugin import find_plugin_classes
    from jasmin_cis.exceptions import ClassNotFoundError
    import logging

    all_classes = find_plugin_classes(parent_class, 'jasmin_cis.col_implementations')
    for cls in all_classes:

        if name is None:
            # Use default
            pass
        else:
            # Method specified directly
            if name == cls.__name__:
                logging.debug("Selected class " +  cls.__name__)
                return cls

    raise ClassNotFoundError("Specified "+parent_class.__name__+" subclass cannot be found")


def get_constraint(method=None):
    '''
    Top level routine for finding the correct Constraint object. This doesn't instantiate the constraint class as it
    may need variables passed to the constructor

    @param method: The constraint method to find - this should be a string which matches the name of one of the subclasses of Constraint
    @return: One of Constraint's subclasses
    '''
    constraint_cls = __get_class(Constraint, method)
    return constraint_cls

def get_kernel(method=None):
    '''
    Top level routine for finding the correct Kernel object.

    @param method: The kernel method to find - this should be a string which matches the name of one of the subclasses of Kernel
    @return: One of Kernel's subclasses
    '''
    kernel_cls = __get_class(Kernel, method)
    return kernel_cls

def get_colocator(method=None):
    '''
    Top level routine for finding the correct Colocator object.

    @param method: The colocate method to find - this should be a string which matches the name of one of the subclasses of Colocator
    @return: One of Colocator's subclasses
    '''
    colocate_cls = __get_class(Colocator, method)
    return colocate_cls



from jasmin_cis.plotting.generic_plot import Generic_Plot

class Contourf_Plot(Generic_Plot):
    def plot(self):
        '''
        Plots a filled contour plot
        '''
        self.contour_plot(True)

    def set_default_axis_label(self, axis):
        return self.set_3daxis_label(axis)

    def format_plot(self):
        self.format_time_axis()
        self.format_3d_plot()



'''
    Module for the UngriddedData class
'''
from netCDF4 import _Variable, Variable
from netcdf import get_data as netcdf_get_data
from hdf_vd import get_data as hdf_vd_get_data, VDS
from pyhdf.SD import SDS
from hdf_sd import get_data as hdf_sd_get_data
import logging


class Metadata(object):

    @classmethod
    def from_CubeMetadata(cls, cube_meta):
        return cls(name='',standard_name=cube_meta.standard_name,long_name=cube_meta.long_name, units=str(cube_meta.units), misc=cube_meta.attributes)


    def __init__(self, name='', standard_name='', long_name='', shape='', units='', range='', factor='', offset='', missing_value='', calendar='', history = '', misc=None):
        self._name = name
        if standard_name:
            self.standard_name = standard_name
        else:
            self.standard_name = Metadata.guess_standard_name(name)
        self.long_name = long_name
        self.shape = shape
        self.units = units
        self.range = range
        self.factor = factor
        self.offset = offset
        self.missing_value = missing_value
        self.calendar = calendar
        self.history = history
        if misc is None:
            self.misc = {}
        else:
            self.misc = misc

    @staticmethod
    def guess_standard_name(name):
        standard_name = name
        if name.lower().startswith('lat'):
            standard_name = 'latitude'
        elif name.lower().startswith('lon'):
            standard_name = 'longitude'
        elif name.lower().startswith('alt') or name.lower() == 'height':
            standard_name = 'altitude'
        return standard_name


# This defines the mappings for each of the ungridded data types to their reading routines, this allows 'lazy loading'
static_mappings = { SDS : hdf_sd_get_data,
                    VDS : hdf_vd_get_data,
                    Variable : netcdf_get_data,
                    _Variable : netcdf_get_data }

class LazyData(object):
    '''
        Wrapper (adaptor) class for the different types of possible ungridded data.
    '''

    def __init__(self, data, metadata, data_retrieval_callback=None):
        '''
        @param data:    The data handler (e.g. SDS instance) for the specific data type, or a numpy array of data
                        This can be a list of data handlers, or a single data handler
        @param metadata: Any associated metadata
        @param data_retrieval_callback: An, optional, method for retrieving data when needed
        '''
        from jasmin_cis.exceptions import InvalidDataTypeError
        from iris.cube import CubeMetadata
        import numpy as np

        self._data_flattened = None

        self.metadata = Metadata.from_CubeMetadata(metadata) if isinstance(metadata, CubeMetadata) else metadata

        if isinstance(data, np.ndarray):
            # If the data input is a numpy array we can just copy it in and ignore the data_manager
            self._data = data
            self._data_manager = None
        else:
            # If the data input wasn't a numpy array we assume it is a data reference (e.g. SDS) and we refer
            #  this as a 'data manager' as it is responsible for getting the actual data.

            self._data = None
            # Although the data can be a list or a single item it's useful to cast it
            #  to a list here to make accessing it consistent
            if isinstance(data, list):
                self._data_manager = data
            else:
                self._data_manager = [ data ]

            if data_retrieval_callback is not None:
                # Use the given data retrieval method
                self.retrieve_raw_data = data_retrieval_callback
            elif self._data_manager[0].__class__ in static_mappings and all([d.__class__ == self._data_manager[0].__class__ for d in self._data_manager ]) :
                # Check that we recognise the data manager and that they are all the same

                # Set the retrieve_raw_data method to it's mapped function name
                self.retrieve_raw_data = static_mappings[self._data_manager[0].__class__]
            else:
                raise InvalidDataTypeError

    def name(self):
        """
            This routine returns the first name property which is not empty out of: _name, standard_name and long_name
                If they are all empty it returns an empty string
        @return: The name of the data object as a string
        """

        for name in [self.metadata._name, self.metadata.standard_name, self.metadata.long_name]:
            if name:
                return name
        return ''

    @property
    def shape(self):
        return self.metadata.shape

    @shape.setter
    def shape(self, shape):
        self.metadata.shape = shape

    @property
    def long_name(self):
        return self.metadata.long_name

    @long_name.setter
    def long_name(self, long_name):
        self.metadata.long_name = long_name

    @property
    def standard_name(self):
        return self.metadata.standard_name

    @standard_name.setter
    def standard_name(self, standard_name):
        self.metadata.standard_name = standard_name

    @property
    def units(self):
        return self.metadata.units

    @units.setter
    def units(self, units):
        self.metadata.units = units

    @property
    def data(self):
        '''
            This is a getter for the data property. It caches the raw data if it has not already been read.
             Throws a MemoryError when reading for the first time if the data is too large.
        '''
        import numpy.ma as ma
        if self._data is None:
            try:
                # If we ere given a list of data managers then we need to concatenate them now...
                self._data=self.retrieve_raw_data(self._data_manager[0])
                if len(self._data_manager) > 1:
                    for manager in self._data_manager[1:]:
                        self._data = ma.concatenate((self._data,self.retrieve_raw_data(manager)),axis=0)
            except MemoryError:
                raise MemoryError(
                    "Failed to read the ungridded data as there was not enough memory available.\n"
                    "Consider freeing up variables or indexing the cube before getting its data.")
        return self._data

    @data.setter
    def data(self, value):
        # TODO remove this - it's only for testing colocation at the moment
        self._data = value

    @property
    def data_flattened(self):
        '''Returns a 1D flattened view (or copy, if necessary) of the data.
        '''
        if self._data_flattened is None:
            data = self.data
            self._data_flattened = data.ravel()
        return self._data_flattened

    def copy_metadata_from(self, other_data):
        '''
            Method to copy the metadata from one UngriddedData/Cube object to another
        '''
        self._coords = other_data.coords()
        self.metadata = other_data._metadata

        #def __getitem__(self, item): pass
        # This method can be overridden to provide the ability to ask for slices of data e.g. UngridedDataObject[012:32.4:5]
        # Actually implementing it would be very tricky as you have to keep track of the data and the coordinates without
        #  necessarily actually reading them


class UngriddedData(LazyData):
    '''
        Wrapper (adaptor) class for the different types of possible ungridded data.
    '''

    def __init__(self, data, metadata, coords, data_retrieval_callback=None):
        '''
        Constructor

        @param data:    The data handler (e.g. SDS instance) for the specific data type, or a numpy array of data
                        This can be a list of data handlers, or a single data handler
        @param metadata: Any associated metadata
        @param coords: A list of the associated Coord objects
        @param data_retrieval_callback: A method for retrieving data when needed
        '''
        from jasmin_cis.data_io.Coord import CoordList, Coord

        super(UngriddedData, self).__init__(data, metadata, data_retrieval_callback)

        if isinstance(coords, list):
            self._coords = CoordList(coords)
        elif isinstance(coords, CoordList):
            self._coords = coords
        elif isinstance(coords, Coord):
            self._coords = CoordList([coords])
        else:
            raise ValueError("Invalid Coords type")
        self.coords_on_grid = False
    
    @property
    def x(self):
        return self.coord(axis='X')

    @property
    def y(self):
        return self.coord(axis='Y')

    @property
    def lat(self):
        return self.coord(standard_name='latitude')

    @property
    def lon(self):
        return self.coord(standard_name='longitude')

    def hyper_point(self, index):
        """

        @param index: The index in the array to find the point for
        @return: A hyperpoint representing the data at that point
        """
        from jasmin_cis.data_io.hyperpoint import HyperPoint
        return HyperPoint(self.coord(standard_name='latitude').data.flat[index],
                          self.coord(standard_name='longitude').data.flat[index],
                          self.coord(standard_name='altitude').data.flat[index],
                          self.coord(standard_name='time').data.flat[index],
                          self.data.flat[index])

    def coords(self, name=None, standard_name=None, long_name=None, attributes=None, axis=None):
        """

        @return: A list of coordinates in this UngriddedData object fitting the given criteria
        """
        return self._coords.get_coords(name, standard_name, long_name, attributes, axis)

    def coord(self, name=None, standard_name=None, long_name=None, attributes=None, axis=None):
        """

        @raise: CoordinateNotFoundError
        @return: A single coord given the same arguments as L(coords).

        """
        return self._coords.get_coord(name, standard_name, long_name, attributes, axis)

    def get_all_points(self):
        """
             Pack a list of coordinates into a list of x, y, z, t points.

        @param coords: A CoordList of Coord objects
        @return: A list of HyperPoints
        """
        import numpy as np
        from hyperpoint import HyperPoint, HyperPointList

        points = HyperPointList()

        logging.info("--> Converting ungridded data to a list of HyperPoints...")

        # We want the data to be flat whether it's masked or not
        data = self.data.flatten()

        # We want the coordinates to be the full length - the way we access masked arrays gives us the indices as if
        #  it were the full length of the array
        data_len = len(data)
        all_coords = self._coords.get_standard_coords(data_len)

        for i, val in enumerate(data):
            points.append(HyperPoint(all_coords[0][i], all_coords[1][i], all_coords[2][i], all_coords[3][i], val))

        return points


    def get_non_masked_points(self):
        """
             Pack a list of coordinates into a list of x, y, z, t points. If the internal data is a masked array then
             only valid (unmasked) points will be returned in the list.

        @param coords: A CoordList of Coord objects
        @return: A list of HyperPoints
        """
        import numpy as np
        from hyperpoint import HyperPoint, HyperPointList

        points = HyperPointList()

        logging.info("--> Converting ungridded data to a list of HyperPoints...")

        # We want the data to be flat whether it's masked or not
        data = self.data.flatten()

        # We want the coordinates to be the full length - the way we access masked arrays gives us the indices as if
        #  it were the full length of the array
        data_len = len(data)
        all_coords = self._coords.get_standard_coords(data_len)

        if isinstance(self.data, np.ma.masked_array):
            # Loop over all of the elements in the array, so that i goes from 0 to len(data)-1, but
            #  only access those entries which the mask is False, that is for which not the mask is True...
            for i, val in enumerate(data):
                if not data.mask[i]:
                    points.append(HyperPoint(all_coords[0][i], all_coords[1][i], all_coords[2][i], all_coords[3][i], val))
        else:
            for i, val in enumerate(data):
                points.append(HyperPoint(all_coords[0][i], all_coords[1][i], all_coords[2][i], all_coords[3][i], val))

        return points


    @classmethod
    def from_points_array(cls, hyperpoints):
        """
         A constuctor for building an UngriddedData object from a list of hyper points
        @param hyperpoints:    A list of HyperPoints
        """
        from jasmin_cis.data_io.Coord import Coord, CoordList
        from jasmin_cis.data_io.hyperpoint import HyperPointList

        if not isinstance(hyperpoints, HyperPointList):
            hyperpoints = HyperPointList(hyperpoints)

        values = hyperpoints.vals
        latitude = hyperpoints.latitudes
        longitude = hyperpoints.longitudes
        altitude = hyperpoints.altitudes
        time = hyperpoints.times

        coords = CoordList( [Coord(latitude, Metadata(standard_name='latitude', units='degrees north')),
                             Coord(longitude, Metadata(standard_name='longitude', units='degrees east')),
                             Coord(altitude, Metadata(standard_name='altitude', units='meters')),
                             Coord(time, Metadata(standard_name='time', units='seconds'))])

        return cls(values, Metadata(), coords)



from abc import ABCMeta, abstractmethod


class SubsetterInterface(object):
    """Interface for subsetter classes.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def subset(self, data, constraint):
        """Subsets the supplied data using the supplied constraint.

        @param data: data to be subsetted
        @param constraint: SubsetConstraint object to be used to subset data
        @return: subsetted data
        """


class SubsetConstraintInterface(object):
    """Interface for subset constraint classes.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def constrain(self, data):
        """Subsets the supplied data.

        @param data: data to be subsetted
        @return: subsetted data
        """



#from jasmin_cis.data_io.products.AProduct import get_data
from glob import glob

from jasmin_cis.data_io.read import read_data


filenames = glob("/home/daniel/Aerosol_CCI_Files/*")
#filenames = [ "/home/daniel/CCI_Files/20080612093821-ESACCI-L2P_AEROSOL-ALL-AATSR_ENVISAT-ORAC_32855-fv02.02.nc" ]
data_object = read_data(filenames, 'AOD550',"Aerosol_CCI")

from plotting.plot import Plotter

Plotter([data_object], plot_type="scatter")


'''
 Module to test the colocation routines
'''
from jasmin_cis.data_io.hyperpoint import HyperPoint
from nose.tools import assert_almost_equal, istest, eq_

@istest
def test_same_point_in_space_and_time_with_points_with_different_values():
    assert(HyperPoint(10.0,50.0,val=14.4).same_point_in_space_and_time(HyperPoint(10.0,50.0,val=15.1)))

@istest
def test_same_point_in_space_with_points_with_different_values():
    assert(HyperPoint(10.0,50.0,10.0,val=14.4).same_point_in_space(HyperPoint(10.0,50.0,10.0,val=15.1)))

@istest
def test_same_point_in_space_with_points_with_different_times():
    assert(HyperPoint(10.0,50.0,10.0,500.0).same_point_in_space(HyperPoint(10.0,50.0,10.0,501.0)))

@istest
def test_same_point_in_time_with_points_with_different_values():
    assert(HyperPoint(10.0,50.0,5.0,500.0,14.4).same_point_in_time(HyperPoint(10.0,50.0,5.0,500.0,15.1)))

@istest
def test_same_point_in_time_with_points_with_different_spatial_coords():
    assert(HyperPoint(10.0,50.0,5.0,500.0).same_point_in_time(HyperPoint(10.0,50.0,10.0,500.0)))

@istest
def can_get_valid_coord_tuple_lat():
    eq_(HyperPoint(10).get_coord_tuple(), [('latitude',10)])

@istest
def can_get_valid_coord_tuple_lat_lon():
    eq_(HyperPoint(10,12).get_coord_tuple(), [('latitude',10), ('longitude',12)])

@istest
def check_furthest_point_on_equator():
    eq_(HyperPoint(0,0).furthest_point_from(),HyperPoint(0,180))

@istest
def check_furthest_point_on_pole():
    eq_(HyperPoint(90,0).furthest_point_from(),HyperPoint(-90,180))

@istest
def check_furthest_point():
    eq_(HyperPoint(51,0).furthest_point_from(),HyperPoint(-51,180))

@istest
def check_dist_between_2d_points_on_equator():
    assert_almost_equal(HyperPoint(0,0).haversine_dist(HyperPoint(0, 1)),111.317,places=3)
    
@istest
def check_dist_between_2d_points_at_pole():
    assert_almost_equal(HyperPoint(90,0).haversine_dist(HyperPoint(90, 1)),0.0000,places=5)
        
@istest
def check_dist_between_opposite_2d_points():
    '''
        The distance between two points on opposite sides of the globe should be half the circumfrence of the globe
    '''
    import math
    R_E = 6378 # Radius of the earth in km
    max_dist = math.pi*R_E
    assert_almost_equal(HyperPoint(0,0).haversine_dist(HyperPoint(0, 180)), max_dist, places=3)
    assert_almost_equal(HyperPoint(90,0).haversine_dist(HyperPoint(-90, 0)), max_dist, places=3)
    assert_almost_equal(HyperPoint(51,0).haversine_dist(HyperPoint(-51, 180)), max_dist, places=3)
    
def check_compdist_returns_true_when_p2_is_closer_than_p1():
    assert(HyperPoint(0,0).compdist(HyperPoint(4,4), HyperPoint(3,3)))
    
def check_compdist_returns_false_when_p1_is_closer_than_p2():
    assert( not HyperPoint(0,0).compdist(HyperPoint(3,3), HyperPoint(4,4)))
    
def check_compdist_returns_false_when_p1_is_the_same_as_p2():
    assert( not HyperPoint(0,0).compdist(HyperPoint(3,3), HyperPoint(3,3)))
    assert( not HyperPoint(0,0).compdist(HyperPoint(-3,3), HyperPoint(3,3)))
    assert( not HyperPoint(0,0).compdist(HyperPoint(3,-3), HyperPoint(3,3)))
    assert( not HyperPoint(0,0).compdist(HyperPoint(-3,-3), HyperPoint(3,3)))
    assert( not HyperPoint(-2.5,0).compdist(HyperPoint(-5,0), HyperPoint(0,0)))
    assert( not HyperPoint(0,-2.5).compdist(HyperPoint(0,-5), HyperPoint(0,0)))
    assert( not HyperPoint(-2.5,-2.5).compdist(HyperPoint(-5,-5), HyperPoint(0,0)))


from collections import namedtuple

class HyperPoint(namedtuple('HyperPoint',['latitude','longitude','altitude','time','val'])):
    '''
     Data type representing a point in space and time. It can contain multiple values which are stored in a list.
      We don't specify a reference time yet but when we do it should probably be here
    '''

    def __new__(cls, lat=None, lon=None, alt=None, t=None, val=None):
        '''
            Small constructor for the HyperPoint named tuple to allow optional arguments
             and set-up value list.
        '''
        import datetime
        from jasmin_cis.time_util import convert_datetime_to_std_time

        # If no value was specified create an empty list, otherwise create a list with one entry
        if val is None or val == []:
            val = []
        else:
            val = [ val ]

        # If t was given as a datetime we need to convert it into our standard time
        if isinstance(t,datetime.datetime):
            t = convert_datetime_to_std_time(t)

        point = super(HyperPoint,cls).__new__(cls,lat,lon,alt,t,val)

        # Store the coordinate tuple for this point in case we need it later
        point.coord_tuple = point.get_coord_tuple()

        return point

    def same_point_in_time(self, other):
        return self.time == other.time

    def same_point_in_space(self, other):
        return ( self.latitude == other.latitude and self.longitude == other.longitude and
                 self.altitude == other.altitude )

    def same_point_in_space_and_time(self, other):
        return ( self.same_point_in_space(other) and self.same_point_in_time(other) )

    def get_coord_tuple(self):
        # This returns a sorted tuple of coorinate names and values. It is sorted to fix an iris bug when doing
        #  linear interpolation. It's linear interpolation routine calls itself recursively and recreates the cube each time,
        #  but it doesn't seem to repopulate derived dimensions. Hence the altitude dimension (which is usually the derived one)
        #  needs to be first in the list of coordinates to interpolate over.
        return sorted([ (x, y) for x, y in self._asdict().items() if y is not None and x != 'val' ])

    def compdist(self,p1,p2):
        '''
            Compares the distance from this point to p1 and p2. Returns True if p2 is closer to self than p1
        '''
        return (self.haversine_dist(p1) > self.haversine_dist(p2))

    def compalt(self,p1,p2):
        '''
            Compares the distance from this point to p1 and p2. Returns True if p2 is closer to self than p1
        '''
        return (self.alt_sep(p1) > self.alt_sep(p2))

    def comptime(self,p1,p2):
        '''
            Compares the distance from this point to p1 and p2. Returns True if p2 is closer to self than p1
        '''
        return (self.time_sep(p1) > self.time_sep(p2))

    def haversine_dist(self,point2):
        '''
            Computes the Haversine distance between two points
        '''
        from jasmin_cis.utils import haversine
        return haversine(self.latitude, self.longitude, point2.latitude, point2.longitude)

    def time_sep(self,point2):
        '''
            Computes the time seperation between two points
        '''
        return abs(self.time - point2.time)

    def alt_sep(self,point2):
        '''
            Computes the height seperation between two points
        '''
        return abs(self.altitude - point2.altitude)

    def furthest_point_from(self):
        '''
            Return a point on the opposite side of the globe from this point
        '''
        furthest_lat = -self.latitude
        if self.longitude > 180:
            furthest_lon = self.longitude - 180.0
        else:
            furthest_lon = self.longitude + 180.0
        return HyperPoint(furthest_lat, furthest_lon, self.altitude, self.time, self.val)


class HyperPointList(list):
    """All the functionality of a standard `list` with added "HyperPoint" context."""

    def __new__(cls, list_of_points=None):
        """
        Given a `list` of HyperPoints, return a HyperPointList instance. This checks that all of the items in the list
         are HyperPoint instances but no other checks. Specifically each Hyper point should have the same shape, that is
         if any point has a coordinate as None, then no other points should have a value for that coordinate. This is not
         imposed here because of the potential overhead but it is assumed by other functions in CIS.

        """
        point_list = list.__new__(cls, list_of_points)

        # Check that all items in the incoming list are HyperPoints. Note that this checking
        # does not guarantee that a HyperPointList instance *always* has just HyperPoints in its list as
        # the append & __getitem__ methods have not been overridden.
        if not all([isinstance(point, HyperPoint) for point in point_list]):
            raise ValueError('All items in list_of_coords must be Coord instances.')
        return point_list

    @property
    def vals(self):
        from numpy import zeros
        n_points = len(self)

        values = zeros(n_points)
        for i, point in enumerate(self):
            values[i] = point.val[0]

        return values

    @property
    def longitudes(self):
        from numpy import zeros

        if self[0].longitude is not None:
            n_points = len(self)

            values = zeros(n_points)
            for i, point in enumerate(self):
                values[i] = point.longitude
        else:
            values = None
        return values

    @property
    def latitudes(self):
        from numpy import zeros

        if self[0].longitude is not None:
            n_points = len(self)

            values = zeros(n_points)
            for i, point in enumerate(self):
                values[i] = point.latitude
        else:
            values = None
        return values

    @property
    def altitudes(self):
        from numpy import zeros

        if self[0].longitude is not None:
            n_points = len(self)

            values = zeros(n_points)
            for i, point in enumerate(self):
                values[i] = point.altitude
        else:
            values = None
        return values

    @property
    def times(self):
        from numpy import zeros

        if self[0].longitude is not None:
            n_points = len(self)

            values = zeros(n_points)
            for i, point in enumerate(self):
                values[i] = point.time
        else:
            values = None
        return values


import logging


def get_all_subclasses(parent_class, mod):
    '''
        This will recursively find subclasses of parent_class in mod.
        The use of importlib allows mod to be of the form package.module
        Take extreme care when changing the function as it has been known to break!
    @param parent_class: The class to find subclasses of
    @param mod: The module to find subclasses in
    @return: A list of subclasses
    '''
    import importlib
    importlib.import_module(mod)
    subclasses = []
    for subclass in parent_class.__subclasses__():
        subclasses += get_all_subclasses(subclass, mod)
    subclasses += parent_class.__subclasses__()
    return subclasses


def find_plugins(plugin_dir, parent_class_name, verbose):
    import logging
    import os, sys

    # if plugin_dir is None, there is no plugin to import, so return an empty list
    if plugin_dir is None:
        return []

    if verbose:
        logging.info("Looking for plugins... ")

    plugin_files = []
    for f in os.listdir(plugin_dir):

        if f.lower().endswith(('.pyc', '__init__.py')): continue

        if f.endswith(".py"):
            plugin_files.append(f[:-3])

    if len(plugin_files) == 0:
        return []

    if verbose:
        logging.info("importing plugin " + str(plugin_files))

    sys.path.insert(0, plugin_dir)

    for plugin in plugin_files:
        module = __import__(plugin)
        classes = [getattr(module, x) for x in dir(module) if isinstance(getattr(module, x), type)]
        product_classes = [ cls for cls in classes if parent_class_name in str(cls.__bases__[0])]

    return product_classes


def find_plugin_classes(parent_class, built_in_module, verbose=True):
    import os
    # find plugin classes, if any
    ENV_PATH = "CIS_PLUGIN_HOME"
    plugin_dir = os.environ.get(ENV_PATH, None)
    plugin_classes = find_plugins(plugin_dir, parent_class.__name__, verbose)

    # find built-in classes, i.e. subclasses of parent_class
    subclasses = get_all_subclasses(parent_class, built_in_module)
    all_classes = plugin_classes + subclasses

    if verbose:
        logging.debug(parent_class.__name__+" subclasses are: " + str(all_classes))

    return all_classes



'''
module to test the hdf4 utility function of hdf_vd.py
'''
from nose.tools import istest, eq_
from jasmin_cis.test.test_files.data import *
import jasmin_cis.data_io.hdf_vd as hdf_vd


@istest
def test_that_can_read_all_variables():
    dict = hdf_vd.get_hdf_VD_file_variables(valid_hdf_vd_file)
    eq_(len(dict),37)
    eq_(True,dict.has_key('Longitude'))

@istest
def test_that_can_get_data():
    vds_dict = hdf_vd.read(valid_hdf_vd_file,'DEM_elevation')
    vds = vds_dict['DEM_elevation']
    data = hdf_vd.get_data(vds)
    eq_(37081,len(data))

@istest
def test_that_can_get_metadata():
    vd = hdf_vd.read(valid_hdf_vd_file,'DEM_elevation')['DEM_elevation']
    metadata = hdf_vd.get_metadata(vd)
    eq_(metadata._name,"DEM_elevation")
    # No standard name, so gets _name by default
    eq_(metadata.standard_name,"DEM_elevation")
    eq_(metadata.long_name,"Digital Elevation Map")
    eq_(metadata.shape,[37081])
    eq_(metadata.units,"meters")
    eq_(metadata.range,[-9999, 8850])
    eq_(metadata.factor,1.0)
    eq_(metadata.offset,0.0)
    eq_(metadata.missing_value,9999)

    vd = hdf_vd.read(valid_hdf_vd_file,'Longitude')['Longitude']
    metadata = hdf_vd.get_metadata(vd)
    eq_(metadata._name,"Longitude")
    eq_(metadata.standard_name,"longitude")
    eq_(metadata.long_name,"Spacecraft Longitude")
    eq_(metadata.shape,[37081])
    eq_(metadata.units,"degrees")
    eq_(metadata.range,[-180.0, 180.0])
    eq_(metadata.factor,1.0)
    eq_(metadata.offset,0.0)
    eq_(metadata.missing_value,None)


from nose import with_setup
from nose.tools import nottest
import os
from jasmin_cis.test.test_util.mock import make_dummy_2d_ungridded_data

tmp_file = "tmp_file"

def delete_tmp_file():
    os.remove(tmp_file)

@nottest
@with_setup(teardown=delete_tmp_file)
def test_write_hdf():
    '''
    Test currently fails. Will not be fixed as there is no use case.
    @return:
    '''
    from jasmin_cis.data_io.write_hdf import write
    obj = make_dummy_2d_ungridded_data()
    write(obj, tmp_file)


from jasmin_cis.plotting.generic_plot import Generic_Plot
import numpy

class Histogram_3D(Generic_Plot):
    valid_histogram_styles = ["bar", "step", "stepfilled"]

    def plot(self):
        '''
        Plots a 3d histogram.
        Requires two data items.
        The first data item is plotted on the x axis, and the second on the y axis
        '''
        from numpy.ma import MaskedArray
        from jasmin_cis.utils import apply_intersection_mask_to_two_arrays
        from jasmin_cis.exceptions import InvalidNumberOfDatagroupsSpecifiedError
        if len(self.packed_data_items) == 2:
            vmin = self.mplkwargs.pop("vmin")
            vmax = self.mplkwargs.pop("vmax")

            xbins = self.calculate_bin_edges("x")
            ybins = self.calculate_bin_edges("y")

            # All bins that have count less than cmin will not be displayed
            cmin = self.plot_args["valrange"].get("vmin", None)
            # All bins that have count more than cmax will not be displayed
            cmax = self.plot_args["valrange"].get("vmax", None)


            # Add y=x line
            min_val = min(self.unpacked_data_items[0]["data"].min(), self.unpacked_data_items[1]["data"].min())
            max_val = max(self.unpacked_data_items[0]["data"].max(), self.unpacked_data_items[1]["data"].max())
            y_equals_x_array = [min_val, max_val]
            self.matplotlib.plot(y_equals_x_array, y_equals_x_array, color = "black", linestyle = "dashed")

            first_data_item, second_data_item = apply_intersection_mask_to_two_arrays(self.unpacked_data_items[0]["data"], self.unpacked_data_items[1]["data"])

            first_data_item = first_data_item.compressed()
            second_data_item = second_data_item.compressed()

            # Use Numpy histogram generator instead of hist2d to allow log scales to be properly plotted
            histogram2ddata = numpy.histogram2d(first_data_item, second_data_item, bins=[xbins, ybins])[0]
            self.matplotlib.pcolor(xbins, ybins, histogram2ddata.T, vmin=cmin, vmax=cmax,
                                   *self.mplargs, **self.mplkwargs)

            self.mplkwargs["vmin"] = vmin
            self.mplkwargs["vmax"] = vmax
        else:
            raise InvalidNumberOfDatagroupsSpecifiedError("Histogram 3D requires two datagroups")

    def unpack_data_items(self):
        return self.unpack_comparative_data()

    def set_plotting_library(self):
        pass

    def calculate_bin_edges(self, axis):
        '''
        Calculates the number of bins for a given axis.
        Uses 10 as the default
        @param axis: The axis to calculate the number of bins for
        @return: The number of bins for the given axis
        '''
        from jasmin_cis.utils import calculate_histogram_bin_edges
        if axis == "x":
            data = self.unpacked_data_items[0]["data"]
        elif axis == "y":
            data = self.unpacked_data_items[1]["data"]

        bin_edges = calculate_histogram_bin_edges(data, axis, self.plot_args[axis + "range"], self.plot_args[axis + "binwidth"], self.plot_args["log" + axis])

        self.plot_args[axis + "range"][axis + "min"] = bin_edges.min()
        self.plot_args[axis + "range"][axis + "max"] = bin_edges.max()

        return bin_edges

    def format_plot(self):
        # We don't format the time axis here as we're only plotting data against data
        self.matplotlib.gca().ticklabel_format(style='sci', scilimits=(-3,3), axis='both')
        self.format_3d_plot()

    def set_default_axis_label(self, axis):
        self.set_default_axis_label_for_comparative_plot(axis)

    def calculate_axis_limits(self, axis, min_val, max_val, step):
        '''
        Calculates the limits for a given axis.
        If the axis is "x" then looks at the data of the first data item
        If the axis is "y" then looks at the data of the second data item
        @param axis: The axis to calculate the limits for
        @return: A dictionary containing the min and max values for the given axis
        '''
        if axis == "x":
            coord_axis = 0
        elif axis == "y":
            coord_axis = 1
        calculated_min, calculated_max = self.calculate_min_and_max_values_of_array_including_case_of_log(axis, self.unpacked_data_items[coord_axis]["data"])

        valrange = {}
        valrange[axis + "min"] = calculated_min if min_val is None else min_val
        valrange[axis + "max"] = calculated_max if max_val is None else max_val
        valrange[axis + "step"] = step

        return valrange

    def apply_axis_limits(self, valrange, axis):
        if len(valrange) != 0:
            if axis == "x":
                step = valrange.pop("xstep", None)
                self.matplotlib.xlim(**valrange)
                if step is not None: valrange["xstep"] = step
            elif axis == "y":
                step = valrange.pop("ystep", None)
                self.matplotlib.ylim(**valrange)
                if step is not None: valrange["ystep"] = step

    def create_legend(self):
        '''
        Overides the create legend method of the Generic Plot as a 3d histogram doesn't need a legend
        '''
        pass

    def add_color_bar(self):
        '''
        Adds a color bar to the plot and labels it as "Frequency"
        '''
        step = self.plot_args["valrange"].get("vstep", None)
        if step is None:
            ticks = None
        else:
            from matplotlib.ticker import MultipleLocator
            ticks = MultipleLocator(step)
        cbar = self.matplotlib.colorbar(orientation = self.plot_args["cbarorient"], ticks = ticks)

        if self.plot_args["cbarlabel"] is None:
            label = "Frequency"
        else:
            label = self.plot_args["cbarlabel"]

        cbar.set_label(label)

    def set_axis_ticks(self, axis, no_of_dims):
        from numpy import arange

        if axis == "x":
            tick_method = self.matplotlib.xticks
            item_index = 0
        elif axis == "y":
            tick_method = self.matplotlib.yticks
            item_index = 1

        if self.plot_args.get(axis + "tickangle", None) is None:
            angle = None
        else:
            angle = self.plot_args[axis + "tickangle"]

        if self.plot_args[axis + "range"].get(axis + "step") is not None:
            step = self.plot_args[axis + "range"][axis + "step"]

            if self.plot_args[axis + "range"].get(axis + "min") is None:
                min_val = self.unpacked_data_items[item_index]["data"].min()
            else:
                min_val = self.plot_args[axis + "range"][axis + "min"]

            if self.plot_args[axis + "range"].get(axis + "max") is None:
                max_val = self.unpacked_data_items[item_index]["data"].max()
            else:
                max_val = self.plot_args[axis + "range"][axis + "max"]

            ticks = arange(min_val, max_val+step, step)

            tick_method(ticks, rotation=angle)
        else:
            tick_method(rotation=angle)

    def drawcoastlines(self):
        pass



import logging
import math
import sys

from iris import cube, coords
from iris.exceptions import IrisError
import iris.unit
import iris.util

import jasmin_cis.cis as cis
import jasmin_cis.exceptions as ex
import jasmin_cis.parse_datetime as parse_datetime
from jasmin_cis.data_io.read import read_data
from jasmin_cis.subsetting.subsetter import Subsetter
from jasmin_cis.subsetting.subset_constraint import (GriddedSubsetConstraint, UngriddedOnGridSubsetConstraint,
                                                     UngriddedSubsetConstraint)
from jasmin_cis.data_io.write_netcdf import add_data_to_file, write_coordinates
from jasmin_cis.cis import __version__


class Subset(object):
    def __init__(self, limits, output_file):
        self._limits = limits
        self._output_file = output_file

    def subset(self, variable, filenames, product):
        # Read the input data - the parser limits the number of data groups to one for this command.
        data = None
        try:
            # Read the data into a data object (either UngriddedData or Iris Cube), concatenating data from
            # the specified files.
            logging.info("Reading data for variable: %s", variable)
            data = read_data(filenames, variable, product)
        except (IrisError, ex.InvalidVariableError) as e:
            raise ex.CISError("There was an error reading in data: \n" + str(e))
        except IOError as e:
            raise ex.CISError("There was an error reading one of the files: \n" + str(e))

        # Set subset constraint type according to the type of the data object.
        if isinstance(data, cube.Cube):
            # Gridded data on Cube
            subset_constraint = GriddedSubsetConstraint()
##         elif data.coords_on_grid:
##             # UngriddedData object with data lying on grid
##             subset_constraint = UngriddedOnGridSubsetConstraint()
        else:
            # Generic ungridded data
            subset_constraint = UngriddedSubsetConstraint()

        self._set_constraint_limits(data, subset_constraint)
        subsetter = Subsetter()
        subset = subsetter.subset(data, subset_constraint)

        if subset is None:
            # Constraints exclude all data.
            raise ex.NoDataInSubsetError("No output created - constraints exclude all data")
        else:
            history = "Subsetted using CIS version " + __version__ + \
                      "\nvariable: " + str(variable) + \
                      "\nfrom files: " + str(filenames) + \
                      "\nusing limits: " + str(subset_constraint)

            if isinstance(subset, cube.Cube):
                subset.add_history(history)
                iris.save(subset, self._output_file)
            else:
                if hasattr(subset.metadata, 'history') and len(subset.metadata.history) > 0:
                    subset.metadata.history += '\n' + history
                else:
                    subset.metadata.history = history
                write_coordinates(subset.coords(), self._output_file)
                add_data_to_file(subset, self._output_file)

    def _set_constraint_limits(self, data, subset_constraint):
        for coord in data.coords():
            # Match user-specified limits with dimensions found in data.
            guessed_axis = self._guess_coord_axis(coord)
            limit = None
            if coord.name() in self._limits:
                limit = self._limits[coord.name()]
            elif guessed_axis is not None:
                if guessed_axis in self._limits:
                    limit = self._limits[guessed_axis]
                elif guessed_axis.lower() in self._limits:
                    limit = self._limits[guessed_axis.lower()]

            if limit is not None:
                if limit.is_time or guessed_axis == 'T':
                    # Ensure that the limits are date/times.
                    dt = parse_datetime.convert_datetime_components_to_datetime(limit.start, True)
                    limit_start = self._convert_datetime_to_coord_unit(coord, dt)
                    dt = parse_datetime.convert_datetime_components_to_datetime(limit.end, False)
                    limit_end = self._convert_datetime_to_coord_unit(coord, dt)
                elif guessed_axis == 'X':
                    # Handle circularity.
                    (limit_start, limit_end) = self._fix_longitude_limits_for_coord(float(limit.start),
                                                                                    float(limit.end),
                                                                                    coord)
                else:
                    # Assume to be a non-time axis.
                    (limit_start, limit_end) = self._fix_non_circular_limits(float(limit.start), float(limit.end))
                subset_constraint.set_limit(coord, limit_start, limit_end)

    @staticmethod
    def _guess_coord_axis(coord):
        """Returns X, Y, Z or T corresponding to longitude, latitude,
        altitude or time respectively if the coordinate can be determined
        to be one of these (based on the standard name only, in this implementation).

        This is intended to be similar to iris.util.guess_coord_axis.
        """
        #TODO Can more be done for ungridded based on units, as with iris.util.guess_coord_axis?
        standard_names = {'longitude': 'X', 'grid_longitude': 'X', 'projection_x_coordinate': 'X',
                          'latitude': 'Y', 'grid_latitude': 'Y', 'projection_y_coordinate': 'Y',
                          'altitude': 'Z', 'time': 'T'}
        if isinstance(coord, iris.coords.Coord):
            guessed_axis = iris.util.guess_coord_axis(coord)
        else:
            guessed_axis = standard_names.get(coord.standard_name.lower())
        return guessed_axis

    @staticmethod
    def _convert_datetime_to_coord_unit(coord, dt):
        """Converts a datetime to be in the unit of a specified Coord.
        """
        if isinstance(coord, iris.coords.Coord):
            # The unit class is then iris.unit.Unit.
            iris_unit = coord.units
        else:
            iris_unit = iris.unit.Unit(coord.units)
        return iris_unit.date2num(dt)

    @staticmethod
    def _fix_longitude_limits_for_coord(limit_start, limit_end, coord):
        #TODO Also check unit as degrees or degrees_east?
        # Would like to do the following but UngriddedData coord does not have circular attribute.
        # if hasattr(coord, 'circular') and coord.circular:
        if isinstance(coord, iris.coords.Coord):
            coord_min = coord.points.min()
            coord_max = coord.points.max()
        else:
            coord_min = coord.data.min()
            coord_max = coord.data.max()

        return Subset._fix_longitude_limits(limit_start, limit_end, coord_min, coord_max)

    @staticmethod
    def _fix_longitude_limits(limit_start, limit_end, coord_min, coord_max):
        new_limits = (limit_start, limit_end)
        do_fix = False
        # Only attempt to modify limits if outside of range of coordinate values.
        if not ((coord_min <= limit_start <= coord_max) and (coord_min <= limit_end <= coord_max)):
            # Determine if -180,180 or 0,360 range is compatible - if neither: do not attempt to fix.
            if (-180.0 <= coord_min < 0.0) and (coord_max <= 180.0):
                do_fix = True
                limit_min = -180.0
            elif (0.0 <= coord_min) and (coord_max <= 360.0):
                do_fix = True
                limit_min = 0.0
        if do_fix:
            new_limits = (Subset._fix_angular_limit(limit_start, limit_min),
                          Subset._fix_angular_limit(limit_end, limit_min))
            logging.info("Angular limits: original: %s  after fix: %s", (limit_start, limit_end), new_limits)

        return new_limits

    @staticmethod
    def _fix_angular_limit(value, range_start):
        """Force an angular variable to be within the 360 range starting from limit_start.
        @param value: value to be fixed to be within the range
        @param range_start: value at start of 360 range assuming -180 <= range_start <= 0.0
        """
        # Narrow to range -360 to 360.
        ret = math.fmod(value, 360.0)
        # Restrict to 360 range starting at range_start.
        if ret < range_start:
            ret += 360.0
        if ret > range_start + 360.0:
            ret -= 360.0
        return ret

    @staticmethod
    def _fix_non_circular_limits(limit_start, limit_end):
        if limit_start <= limit_end:
            new_limits = (limit_start, limit_end)
        else:
            new_limits = (limit_end, limit_start)
            logging.info("Real limits: original: %s  after fix: %s", (limit_start, limit_end), new_limits)

        return new_limits



import datetime
import re


def _parse_datetime(dt_string):
    """Parse a date/time string.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T. 
    @param dt_string: String to parse
    @return: list of datetime components
    @raise ValueError: if the string cannot be parsed as a date/time
    """
    # Separate date and time at first character that is one of 'T', ' ' or ':'.
    match = re.match(r'(?P<date>[^T :]+)(?:[T :])(?P<time>.+)$', dt_string)
    if match is not None:
        date_str = match.group('date')
        time_components = [int(x) for x in match.group('time').split(':')]
    else:
        date_str = dt_string
        time_components = []

    date_components = [int(x) for x in date_str.split('-')]

    dt_components = list(date_components)
    if ((len(date_components) > 3) or (len(time_components) > 3) or
            (len(date_components) < 3) and (len(time_components) > 0)):
        raise ValueError()
    else:
        dt_components.extend(time_components)

    # Check that the components are valid (assuming month and/or date is 1 if missing).
    tmp_components = list(dt_components)
    if len(tmp_components) < 3:
        tmp_components.extend([1] * (3 - len(tmp_components)))
    dt = datetime.datetime(*tmp_components)

    return dt_components


def parse_datetime(dt_string, name, parser):
    """Parse a date/time string from the command line, reporting parse errors.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    @param dt_string: String to parse
    @param name:      A description of the argument used for error messages
    @param parser:    The parser used to report errors
    @return: datetime value
    """
    try:
        dt = _parse_datetime(dt_string)
    except ValueError:
        parser.error("'" + dt_string + "' is not a valid " + name)
        dt = None
    return dt


def parse_as_number_or_datetime(in_string, name, parser):
    """Parse a string as a number from the command line, or if that fails, as a datetime, reporting parse errors.

    The string should be in an ISO 8601 format except that the date and time
    parts may be separated by a space or colon instead of T.
    @param in_string: String to parse
    @param name:      A description of the argument used for error messages
    @param parser:    The parser used to report errors
    @return: int, float or datetime value
    """
    try:
        ret = int(in_string)
    except ValueError:
        try:
            ret = float(in_string)
        except ValueError:
            try:
                ret = _parse_datetime(in_string)
            except ValueError:
                parser.error("'" + in_string + "' is not a valid " + name)
                ret = None
    return ret


def convert_datetime_components_to_datetime(dt_components, is_lower_limit):
    """Converts date and time components: year, month, day, hours, minutes, seconds to a datetime object for use
    as a limit.

    Components beyond the year are defaulted if absent, taking minimum or maximum values if the value of
    is_lower_limit is True or False respectively.
    @param dt_components: list of date/time components as integers in the order:year, month, day, hours, minutes,
        seconds. At least the year must be specified. If only the year is specified it may be passed as an integer.
    @param is_lower_limit: If True, use minimum value for missing date/time components, otherwise use maximum.
    """
    YEAR_INDEX = 0
    DAY_INDEX = 2

    if isinstance(dt_components, list):
        all_components = list(dt_components)
    else:
        all_components = [dt_components]

    if is_lower_limit:
        default_limits = [None, 1, 1, 0, 0, 0]
    else:
        default_limits = [None, 12, None, 23, 59, 59]

    # Fill in the trailing fields with default values.
    for idx in range(len(all_components), len(default_limits)):
        if idx == YEAR_INDEX:
            # Don't default the year - this is the minimum that must be specified.
            pass
        elif idx == DAY_INDEX and not is_lower_limit:
            # Set day to last day of month.
            all_components.append(find_last_day_of_month(*all_components))
        else:
            all_components.append(default_limits[idx])
    return datetime.datetime(*all_components)


def find_last_day_of_month(year, month):
    """Finds the last day of a month.

    @param year: year as integer
    @param month: month as integer
    @return: last day of month as integer
    """
    next_month = datetime.datetime(year, month, 28) + datetime.timedelta(days=4)
    last_date = next_month - datetime.timedelta(days=next_month.day)
    return last_date.day




import numpy as np
from pyhdf.error import HDF4Error
import hdf

def dictkeysappend(dict1,dict2,axis=0):
    ''' Appends data in dict2 to already existing keys in a dictionary dict1. If that key does not exist, it is created'''
    for key in dict2.keys():
        try:
            dict1[key] = np.concatenate((dict1[key],dict2[key]),axis=axis)
        except KeyError:
            dict1[key] = dict2[key]
    return dict1


def readin_MODIS_L2(filenames,sds,outdata = None):
    '''Reads in single day of MODIS L2 data'''

    if outdata == None:
        outdata = {}

    names = sds + ['Latitude','Longitude']

    for file in filenames:
        #print file
        try:
            print "reading: " + file
            data_dict = hdf.read_hdf4(file,names,vdata=None)
            dictkeysappend(outdata,data_dict)
        except HDF4Error as e:
            print e.message, 'in readin_MODIS_L2', file
    return outdata


def field_interpolate(data,factor=5):
    '''Interpolates the given 2D field by the factor, edge pixels are defined by the ones in the centre, odd factords only!'''
    output = np.zeros((factor*data.shape[0],factor*data.shape[1]))*np.nan
    output[int(factor/2)::factor,int(factor/2)::factor] = data
    for i in range(1,factor+1):
        output[(int(factor/2)+i):(-1*factor/2+1):factor,:] = i*((output[int(factor/2)+factor::factor,:]-output[int(factor/2):(-1*factor):factor,:])
                                                                /float(factor))+output[int(factor/2):(-1*factor):factor,:]
    for i in range(1,factor+1):
        output[:,(int(factor/2)+i):(-1*factor/2+1):factor] = i*((output[:,int(factor/2)+factor::factor]-output[:,int(factor/2):(-1*factor):factor])
                                                                /float(factor))+output[:,int(factor/2):(-1*factor):factor]

    return output


if __name__ ==  '__main__':

    import matplotlib.pyplot as plt
    from mpl_toolkits.basemap import Basemap
    from glob import glob

    # get a list of files to read
    folder = "/home/david/Data"
    filenames = glob(folder + "/*" + "MYD06_L2" + "*.hdf")
    filenames = np.sort(filenames)

    #Scan_Start_Time is the time of the MODIS observations - it is not read in by default as
    # the joint L2 product (MYDATML2) omits it.  However, the scan time can be approximated
    # by the time in the filename
    #The is currently only collection 5.1 (col='51') on cloud for MYD06 for a limited area
    # around the Barbados Cloud Observatory
    L2_data = readin_MODIS_L2(filenames,['Cloud_Water_Path','Cloud_Top_Pressure_Day','Scan_Start_Time'])

    #The latitude and longitude are on a 5km grid, whereas the data is on a 1km grid.  MODIS
    # data comes at several different scales - the cloud optical properties retrieval is done
    # at 1km, some retrievals are at 5km and the aerosol retrieval is at 10km.  If relevant,
    # 5km and 10km lat/lon are usually included - to get 1km lat/lon you have to interpolate.
    # This is a very simple interpolation linear scheme which could do with some improvement.
    latitude_1km = field_interpolate(L2_data['Latitude'])
    longitude_1km = field_interpolate(L2_data['Longitude'])
    #latitude_1km = L2_data['Latitude']
    #longitude_1km = L2_data['Longitude']


    #The last four bins in the across scan direction are truncated (According to P. Hubanks
    # at NASA), I can't find this mentionned anywhere in the documentation though, and it is probably
    # not true for all datasets
    plt_lwp = L2_data['Cloud_Water_Path'][:,:-4]
    plt_lwp = np.ma.array(plt_lwp,mask=np.isnan(plt_lwp))

    m = Basemap(-75,0,-45,35,resolution='i')
    plt.subplot(221)
    m.pcolormesh(longitude_1km,latitude_1km,plt_lwp,vmax=600)
    m.drawcoastlines()
    plt.xticks([-75,-65,-55,-45],[-75,-65,-55,-45])
    plt.yticks([0,5,10,15,20,25,30,35],[0,5,10,15,20,25,30,35])
    plt.xlabel('Longitude')
    plt.ylabel('Latitude')
    plt.title('MYD06 - Cloud Water Path (1km)')
    cb = plt.colorbar(shrink=0.8)
    cb.set_label('gm $^{-2}$')


    plt.show()



#from jasmin_cis.data_io.products.AProduct import get_data
from glob import glob

from jasmin_cis.data_io.read import read_data


filenames = glob("/home/daniel/Downloads/Fixed_CCI_Files_From_Gareth/*")
#filenames = [ "/home/daniel/CCI_Files/Modis_L2/20080620000000-ESACCI-L2_CLOUD-CLD_PRODUCTS-MODIS-AQUA-fv1.0.nc" ]
#filenames = [ "/home/daniel/Downloads/Fixed_CCI_Files_From_Gareth/20080620073500-ESACCI-L2_CLOUD-CLD_PRODUCTS-MODIS-AQUA-fv1.0.nc"]
data_object = read_data(filenames, 'stemp', "Cloud_CCI")

from plotting.plot import Plotter

Plotter([data_object], valrange = {"vmin" :260, "vmax" : 340}, plot_type="scatter", itemwidth=1)


"""
Utilities for converting time units
"""
import numpy as np
from iris.unit import Unit

cis_standard_time_unit = Unit('days since 1600-01-01 00:00:00',calendar='gregorian')

def parse_datetimestr_to_std_time(s):
    import dateutil.parser as du
    return cis_standard_time_unit.date2num(du.parse(s))


def parse_datetimestr_to_std_time_array(string_time_array):
    return convert_numpy_array(string_time_array, 'float64', parse_datetimestr_to_std_time)


def parse_datetimestr_delta_to_float_days(s):
    """
    parsing "1y2m3d4H5M6S" into a fractional day
    @param s: string to be parsed
    @return: a float representation of a day
    """
    import re
    from datetime import timedelta

    tokens = re.findall('[0-9]*[ymdHMS]',s)
    sec = 1.0/(24.0*60.0*60.0) # Conversion from sec to day

    days = hours = minutes = seconds = 0
    for token in tokens:

        val = int(token[:-1])
        if token[-1:] == "y":
            days += val*365.2425
        elif token[-1:] == "m":
            days += val*365.2425/12.0
        elif token[-1:] == "d":
            days += val
        elif token[-1:] == "H":
            hours = val
        elif token[-1:] == "M":
            minutes = val
        elif token[-1:] == "S":
            seconds = val
        else:
            raise ValueError("Invalid time delta format. Must be '1y2m3d4H5M6S'")

    td = timedelta(days=days,hours=hours,minutes=minutes,seconds=seconds)

    return td.total_seconds()*sec


def calculate_mid_time(t1, t2):
    '''
        Find the mid time between two times expressed as floats
    @param t1: a time represented as a float
    @param t2: a time in the same representation as t1
    @return: a float representing the time between t1 and t2
    '''
    return t1 + (t2 - t1)/2.0


def convert_time_since_to_std_time(time_array, units):
    from netcdftime import _dateparse
    import datetime
    units, utc_offset, dt = _dateparse(units)
    dt = datetime.datetime(*dt.timetuple()[0:6])
    if units.lower() == 'days':
        new_array = convert_numpy_array(time_array, 'float64', convert_days_since_to_std_time, dt)
    elif units.lower().startswith('sec'):
        new_array = convert_numpy_array(time_array, 'float64', convert_sec_since_to_std_time, dt)
    else:
        raise NotImplementedError('Tried to convert a time with unrecognised units')
    return new_array


def convert_sec_since_to_std_time_array(tai_time_array, ref):
    return convert_numpy_array(tai_time_array, 'float64', convert_sec_since_to_std_time, ref)


def convert_sec_since_to_std_time(seconds, ref):
    '''
        Convert a number of seconds since a given reference datetime to a number of days since our standard time.
        This in principle could avoid the intermediate step converting to a datetime object except we don't know which
         calander the reference is on, e.g. it could be a 360 day calendar
    @param seconds:
    @param ref:
    @return:
    '''
    from datetime import timedelta
    return cis_standard_time_unit.date2num(timedelta(seconds=float(seconds)) + ref)


def convert_days_since_to_std_time(days, ref):
    from datetime import timedelta
    return cis_standard_time_unit.date2num(timedelta(days=float(days)) + ref)


def convert_std_time_to_datetime(std_time):
    return cis_standard_time_unit.num2date(std_time)


def convert_datetime_to_std_time(dt):
    return cis_standard_time_unit.date2num(dt)


def convert_julian_date_to_std_time_array(julian_time_array, calender='standard'):
    return convert_numpy_array(julian_time_array, 'float64', convert_julian_date_to_std_time, calender)


def convert_julian_date_to_std_time(julian_date, calender='standard'):
    from iris.unit import julian_day2date
    return cis_standard_time_unit.date2num(julian_day2date(julian_date, calender))


def convert_obj_to_standard_date_array(time_array):
    return convert_numpy_array(time_array, 'float64', convert_datetime_to_std_time)


def convert_masked_array_type(masked_array, new_type, operation, *args, **kwargs):
    from numpy.ma import getmaskarray
    converted_time = np.ma.array(np.zeros(masked_array.shape, dtype=new_type),
                                 mask=masked_array.mask)

    masks = getmaskarray(masked_array)
    for i, val in np.ndenumerate(masked_array):
        if not masks[i]:
            converted_time[i] = operation(val, *args, **kwargs)
    return converted_time


def convert_array_type(array, new_type, operation, *args, **kwargs):
    converted_time = np.zeros(array.shape, dtype=new_type)
    for i, val in np.ndenumerate(array):
        converted_time[i] = operation(val, *args, **kwargs)
    return converted_time


def convert_numpy_array(array, new_type, operation, *args, **kwargs):
    if isinstance(array,np.ma.MaskedArray):
        new_array = convert_masked_array_type(array, new_type, operation, *args, **kwargs)
    else:
        new_array = convert_array_type(array, new_type, operation, *args, **kwargs)
    return new_array


def convert_cube_time_coord_to_standard_time(cube):
    # Get the current time coordinate and it's data dimension
    t_coord = cube.coord(standard_name='time')
    data_dim = cube.coord_dims(t_coord)

    # And remove it from the cube
    cube.remove_coord(t_coord)

    # Convert the raw time numbers to our 'standard' time
    new_datetimes = convert_numpy_array(t_coord.points,'O',t_coord.units.num2date)
    new_datetime_nums = convert_obj_to_standard_date_array(new_datetimes)

    # Create a new time coordinate by copying the old one, but using our new points and units
    new_time_coord = t_coord
    new_time_coord.points = new_datetime_nums
    new_time_coord.units = cis_standard_time_unit

    # And add the new coordinate back into the cube
    cube.add_dim_coord(new_time_coord, data_dim)

    return cube



from ungridded_data import LazyData
import logging, collections

class Coord(LazyData):

    def __init__(self, data, metadata, axis=''):
        """

        @param data:
        @param metadata:
        @param axis: A string label for the axis, e.g. 'X', 'Y', 'Z', or 'T'
        @return:
        """
        super(Coord, self).__init__(data, metadata)
        self.axis = axis.upper()

    def __eq__(self, other):
        return other.metadata.standard_name == self.metadata.standard_name and self.metadata.standard_name != ''

    def convert_julian_to_std_time(self, calender='standard'):
        from jasmin_cis.time_util import convert_julian_date_to_std_time_array, cis_standard_time_unit
        #if not self.units.startswith("Julian Date"): raise ValueError("Time units must be Julian Date for conversion to an Object")
        self._data = convert_julian_date_to_std_time_array(self.data, calender)
        self.units = str(cis_standard_time_unit)
        self.metadata.calendar = cis_standard_time_unit.calendar

    def convert_TAI_time_to_std_time(self, ref):
        from jasmin_cis.time_util import convert_sec_since_to_std_time_array, cis_standard_time_unit
        self._data = convert_sec_since_to_std_time_array(self.data, ref)
        self.units = str(cis_standard_time_unit)
        self.metadata.calendar = cis_standard_time_unit.calendar

    def convert_to_std_time(self):
        from jasmin_cis.time_util import convert_time_since_to_std_time, cis_standard_time_unit
        self._data = convert_time_since_to_std_time(self.data, self.units)
        self.units = str(cis_standard_time_unit)
        self.metadata.calendar = cis_standard_time_unit.calendar

    def convert_datetime_to_standard_time(self):
        from jasmin_cis.time_util import convert_obj_to_standard_date_array, cis_standard_time_unit
        self._data = convert_obj_to_standard_date_array(self.data)
        self.units = str(cis_standard_time_unit)
        self.metadata.calendar = cis_standard_time_unit.calendar


class CoordList(list):
    """All the functionality of a standard `list` with added "Coord" context."""

    def __new__(cls, list_of_coords=None):
        """
        Given a `list` of Coords, return a CoordList instance.

        """
        coord_list = list.__new__(cls, list_of_coords)

        # Check that all items in the incoming list are coords. Note that this checking
        # does not guarantee that a CoordList instance *always* has just coords in its list as
        # the append & __getitem__ methods have not been overridden.
        if not all([isinstance(coord, Coord) for coord in coord_list]):
            raise ValueError('All items in list_of_coords must be Coord instances.')

        return coord_list

    def append(self, other):
        """
            Safely add a new coordinate object to the list, this checks for a unique axis and standard_name
        @param other: Other coord to add
        @raise: DuplicateCoordinateError
        @return:
        """
        from jasmin_cis.exceptions import DuplicateCoordinateError
        if any([ other == item for item in self ]):
            raise DuplicateCoordinateError()
        super(CoordList, self).append(other)

    def get_coords(self, name=None, standard_name=None, long_name=None, attributes=None, axis=None):
        """
        Return a list of coordinates in this UngriddedData object fitting the given criteria. This is deliberately very
         similar to Cube.coords() to maintain a similar interface and because the functionality is similar. There
          is no distrinction between dimension coordinates and auxilliary coordinates here though.

        @param name:  The standard name or long name or default name of the desired coordinate.
            If None, does not check for name. Also see, :attr:`Cube.name`.
        @param standard_name: The CF standard name of the desired coordinate. If None, does not check for standard name.
        @param long_name: An unconstrained description of the coordinate. If None, does not check for long_name.
        @param attributes: A dictionary of attributes desired on the coordinates. If None, does not check for attributes.
        @param axis: The desired coordinate axis, see :func:`iris.util.guess_coord_axis`. If None, does not check for axis.
            Accepts the values 'X', 'Y', 'Z' and 'T' (case-insensitive).

        @return: A list of coordinates in this UngriddedData object fitting the given criteria
        """
        from collections import Mapping
        coords = self

        if name is not None:
            coords = filter(lambda coord_: coord_.name() == name, coords)

        if standard_name is not None:
            coords = filter(lambda coord_: coord_.standard_name == standard_name, coords)

        if long_name is not None:
            coords = filter(lambda coord_: coord_.long_name == long_name, coords)

        if axis is not None:
            axis = axis.upper()
            coords = filter(lambda coord_: coord_.axis == axis, coords)

        if attributes is not None:
            if not isinstance(attributes, Mapping):
                raise ValueError('The attributes keyword was expecting a dictionary type, but got a %s instead.' % type(attributes))
            filter_func = lambda coord_: all(k in coord_.attributes and coord_.attributes[k] == v for k, v in attributes.iteritems())
            coords = filter(filter_func, coords)

        return coords

    def get_coord(self, name=None, standard_name=None, long_name=None, attributes=None, axis=None):
        """
        Return a single coord given the same arguments as L(coords). If the arguments given do not result in precisely
         1 coordinate being matched, a CoordinateNotFoundError is raised.

        @raise: CoordinateNotFoundError
        @return: A single coord given the same arguments as L(coords).

        """
        from jasmin_cis.exceptions import CoordinateNotFoundError
        coords = self.get_coords(name=name, standard_name=standard_name, long_name=long_name, attributes=attributes, axis=axis)
        if len(coords) == 0:  # If we found none by name, try with standard name only
            coords = self.get_coords(standard_name=name)

        if len(coords) > 1:
            msg = 'Expected to find exactly 1 coordinate, but found %s. They were: %s.'\
                  % (len(coords), ', '.join(coord.name() for coord in coords))
            raise CoordinateNotFoundError(msg)
        elif len(coords) == 0:
            bad_name = name or standard_name or long_name or axis or ''
            msg = 'Expected to find exactly 1 %s coordinate, but found none.' % bad_name
            raise CoordinateNotFoundError(msg)

        return coords[0]

    def get_coordinates_points(self):
        """
             Pack a list of coordinates into a list of x, y, z, t points to be passed to Colocator
        @param coords: A CoordList of Coord objects
        @return: A list of HyperPoints
        """
        from jasmin_cis.exceptions import CoordinateNotFoundError
        from hyperpoint import HyperPoint, HyperPointList
        points = HyperPointList()

        logging.info("Converting coordinates to a list of HyperPoints")

        # Pick the length of the 0th element in the list as our length - all coordinates should have the same length anyway
        data_len = len(self[0].data.flatten())

        all_coords = self.get_standard_coords(data_len)

        for x in xrange(data_len):
            points.append(HyperPoint(all_coords[0][x], all_coords[1][x], all_coords[2][x], all_coords[3][x]))

        return points

    def get_standard_coords(self, data_len):
        from jasmin_cis.exceptions import CoordinateNotFoundError

        empty_data = [None for i in xrange(data_len)]

        try:
            lat = self.get_coord(standard_name='latitude').data.flatten()
        except CoordinateNotFoundError:
            lat = empty_data
        try:
            lon = self.get_coord(standard_name='longitude').data.flatten()
        except CoordinateNotFoundError:
            lon = empty_data
        try:
            alt = self.get_coord(standard_name='altitude').data.flatten()
        except CoordinateNotFoundError:
            alt = empty_data
        try:
            time = self.get_coord(standard_name='time').data.flatten()
        except CoordinateNotFoundError:
            time = empty_data

        return CoordList([lat, lon, alt, time ])




from jasmin_cis.data_io.read import read_data, read_coordinates

filename = "/home/duncan/jasmin_cis/jasmin_cis/test/test_files/CAL_LID_L2_05kmAPro-Prov-V3-01.2009-12-31T23-36-08ZN.hdf"
filenames = [filename]


from jasmin_cis.exceptions import InvalidVariableError
from jasmin_cis.data_io.hdf import get_hdf4_file_metadata
from jasmin_cis.data_io.hdf_vd import get_data, VDS


my_data = get_data(VDS(filename,'Production_Script'), first_record=True)

print my_data

# metadata_dict = get_hdf4_file_metadata(filename)
#
# for item,val in metadata_dict.iteritems():
#     print item, val
#print metadata_dict

#data_obj = read_data(filenames, "Particulate_Depolarization_Ratio_Profile_532", "Caliop")



#print data_obj.metadata.shape
#print data_obj.data
# Eds code from jasmin_cis.test.harness.hdf import read_hdf4
# Eds code datad = read_hdf4(filename) #DELETE ME
#from plotting.plot import Plotter

#Plotter([data_obj],plot_type="scatter", yrange={"ymin":0}, valrange={"vmin":0.0, "vmax":1.0})
#Plotter([data_obj],plot_type="scatter")




'''
Module that contains various strings that are used in tests
'''

import os

def make_pathname(filename):
    return os.path.join(os.path.dirname(__file__), filename)

valid_hdf_vd_file = make_pathname("2008045004519_09563_CS_2C-PRECIP-COLUMN_GRANULE_P_R04_E02.hdf")
valid_hdf_sd_file = make_pathname("MOD04_L2.A2010001.2255.005.2010005215814.hdf")

valid_modis_l2_filename = make_pathname("MOD04_L2.A2010001.2255.005.2010005215814.hdf")
valid_modis_l2_variable = 'Asymmetry_Factor_Average_Ocean'

valid_modis_l3_filename = make_pathname("MOD08_E3.A2010009.005.2010026072315.hdf")
valid_modis_l3_variable = "Pressure_Level"

valid_cloudsat_RVOD_file = make_pathname("2007180125457_06221_CS_2B-CWC-RVOD_GRANULE_P_R04_E02.hdf")
valid_cloudsat_RVOD_variable = "RVOD_liq_water_content"

valid_aerosol_cci_filename = make_pathname("20080612093821-ESACCI-L2P_AEROSOL-ALL-AATSR_ENVISAT-ORAC_32855-fv02.02.nc")
valid_aerosol_cci_variable = "AOD550"
valid_cloud_cci_filename = make_pathname("20080620072500-ESACCI-L2_CLOUD-CLD_PRODUCTS-MODIS-AQUA-fv1.0.nc")
valid_cloud_cci_variable = "stemp"

valid_aeronet_filename = make_pathname("920801_091128_Agoufou_small.lev20")
another_valid_aeronet_filename = make_pathname("920801_091128_Abracos_Hill_small.lev20")
valid_aeronet_variable = "AOT_440"

valid_caliop_l2_filename = make_pathname("CAL_LID_L2_05kmAPro-Prov-V3-01.2009-12-31T23-36-08ZN.hdf")
valid_caliop_l2_variable = "Tropopause_Temperature"

valid_caliop_l1_filename = make_pathname("CAL_LID_L1-ValStage1-V3-01.2009-12-31T23-36-08ZN.hdf")
valid_second_caliop_l1_filename = make_pathname("CAL_LID_L1-ValStage1-V3-01.2010-01-01T00-22-28ZD.hdf")
valid_caliop_l1_variable = "Tropopause_Temperature"

valid_xenida_filename = make_pathname("xenida.pah9440.nc")
valid_xenida_variable = "mass_fraction_of_cloud_liquid_water_in_air"

valid_1d_filename = make_pathname("xglnwa.pm.k8dec-k9nov.vprof.tm.nc")
valid_1d_vairable = "q"
valid_2d_filename = make_pathname("xglnwa.pm.k8dec-k9nov.col.tm.nc")

valid_cis_col_file = make_pathname("cis-col-Agoufou_seasonal_average.nc")

valid_ascii_filename = make_pathname("my_dummy_points_with_values.txt")
ascii_filename_with_no_values = make_pathname("my_dummy_points.txt")
valid_ascii_variable = 0

valid_NCAR_NetCDF_RAF_filename = make_pathname("RF04.20090114.192600_035100.PNI.nc")
valid_NCAR_NetCDF_RAF_variable = 'ATX'

valid_zonal_time_mean_CMIP5_filename = \
    make_pathname('rsutcs_Amon_HadGEM2-A_sstClim_r1i1p1_185912-188911.CMIP5.tm.zm.nc')
valid_zonal_time_mean_CMIP5_variable = "toa_outgoing_shortwave_flux_assuming_clear_sky"

test_directory = make_pathname('test_directory_for_parser')
test_directory_file1 = test_directory + "/test_file_for_parser_1"
test_directory_file2 = test_directory + "/test_file_for_parser_2"
test_directory_file3 = test_directory + "/test_file_for_parser_3"
test_directory_files = [test_directory_file1, test_directory_file2, test_directory_file3]

large_15GB_file_filename = "/run/media/daniel/Storage/xglnwa.pm.k8dec-k9nov.nc"
invalid_filename = "invalidfilename"
non_netcdf_file = make_pathname("notanetcdffile")
file_without_read_permissions = make_pathname("Unreadable Folder/xglnwa.pm.k8dec-k9nov.vprof.tm.nc")
netcdf_file_with_incorrect_file_extension = make_pathname("xglnwa.pm.k8dec-k9nov.vprof.tm.waldm")
non_netcdf_file_with_netcdf_file_extension = make_pathname("notarealnetcdffile.nc")
not1Dvariable = "u" 
valid_variable_in_valid_filename = "rain"
out_filename = "cube_image.png"
valid_colour = "green"
invalid_colour = "greenn"
valid_width = 40
valid_line_style = "dashed"
valid_colour_map = "RdBu"
invalid_colour_map = "invalid"
valid_font_size = 10
valid_height = 40
valid_ymin = 0
valid_ymax = 1


from jasmin_cis.plotting.generic_plot import Generic_Plot

class Line_Plot(Generic_Plot):
    line_styles = ["solid", "dashed", "dashdot", "dotted"]

    def plot(self):
        '''
        Plots one or many line graphs
        '''
        self.mplkwargs["linewidth"] = self.plot_args.get("itemwidth", 1) if self.plot_args.get("itemwidth", 1) is not None else 1

        self.mplkwargs.pop("vmax", None)
        self.mplkwargs.pop("vmin", None)

        for i, unpacked_data_item in enumerate(self.unpacked_data_items):
            datafile = self.plot_args["datagroups"][i]
            if datafile["itemstyle"]:
                if datafile["itemstyle"] in Line_Plot.line_styles:
                    self.mplkwargs["linestyle"] = datafile["itemstyle"]
                else:
                    from jasmin_cis.exceptions import InvalidLineStyleError
                    raise InvalidLineStyleError("'" + datafile["itemstyle"] + "' is not a valid line style, please use one of: " + str(self.line_styles))
            else:
                self.mplkwargs.pop("linestyle", None)

            if datafile["color"]:
                self.mplkwargs["color"] = datafile["color"]
            else:
                self.mplkwargs.pop("color", None)

            self.matplotlib.plot(unpacked_data_item["x"], unpacked_data_item["data"], *self.mplargs, **self.mplkwargs )

    def format_plot(self):
        self.format_time_axis()
        self.format_2d_plot()

    def set_default_axis_label(self, axis):
        import jasmin_cis.exceptions as cisex
        import iris.exceptions as irisex
        axis = axis.lower()
        axislabel = axis + "label"

        if self.plot_args[axislabel] is None:
            try:
                units = self.packed_data_items[0].coord(name=self.plot_args[axis + "_variable"]).units
            except (cisex.CoordinateNotFoundError, irisex.CoordinateNotFoundError):
                units = self.packed_data_items[0].units

            if len(self.packed_data_items) == 1:
                try:
                    name = self.packed_data_items[0].coord(name=self.plot_args[axis + "_variable"]).name()
                except (cisex.CoordinateNotFoundError, irisex.CoordinateNotFoundError):
                    name = self.packed_data_items[0].name()
                # only 1 data to plot, display
                self.plot_args[axislabel] = name + " " + self.format_units(units)
            else:
                # if more than 1 data, legend will tell us what the name is. so just displaying units
                self.plot_args[axislabel] = self.format_units(units)

    def calculate_axis_limits(self, axis, min_val, max_val, step):
        if axis == "x":
            coord_axis = "x"
        elif axis == "y":
            coord_axis = "data"
        calculated_min, calculated_max = self.calculate_min_and_max_values_of_array_including_case_of_log(axis, self.unpacked_data_items[0][coord_axis])

        valrange = {}
        valrange[axis + "min"] = calculated_min if min_val is None else min_val
        valrange[axis + "max"] = calculated_max if max_val is None else max_val
        valrange[axis + "step"] = step

        return valrange



'''
Class for plotting graphs.
Also contains a dictionary for the valid plot types
All plot types need to be imported and added to the plot_types dictionary in order to be used.
'''
from jasmin_cis.plotting.contour_plot import Contour_Plot
from jasmin_cis.plotting.contourf_plot import Contourf_Plot
from jasmin_cis.plotting.heatmap import Heatmap
from jasmin_cis.plotting.line_plot import Line_Plot
from jasmin_cis.plotting.scatter_overlay import Scatter_Overlay
from jasmin_cis.plotting.scatter_plot import Scatter_Plot
from jasmin_cis.plotting.comparative_scatter import Comparative_Scatter
from jasmin_cis.plotting.histogram2d import Histogram_2D
from jasmin_cis.plotting.histogram3d import Histogram_3D
import matplotlib.pyplot as mpl

plot_options = { 'title' : mpl.title,
                 'xlabel' : mpl.xlabel,
                 'ylabel' : mpl.ylabel,
                 'fontsize' : mpl.rcParams.update }

class Plotter(object):

    plot_types = {"contour" : Contour_Plot,
                  "contourf" : Contourf_Plot,
                  "heatmap" : Heatmap,
                  "line": Line_Plot,
                  "scatteroverlay" : Scatter_Overlay,
                  "scatter" : Scatter_Plot,
                  "comparativescatter" : Comparative_Scatter,
                  "histogram2d" : Histogram_2D,
                  "histogram3d" : Histogram_3D}

    def __init__(self, packed_data_items, plot_type = None, out_filename = None, *mplargs, **mplkwargs):
        '''
        Constructor for the plotter

        @param packed_data_items: A list of packed (i.e. Iris cubes or UngriddedData objects) data items to be plotted
        @param plot_type: The plot type to be used, as a string
        @param out_filename: The filename of the file to save the plot to. Optional. Various file extensions can be used, with png being the default
        @param mplargs: Any other arguments received from the parser
        @param mplkwargs: Any other keyword arguments received from the plotter
        '''
        mplkwargs.pop("xmin", None)
        mplkwargs.pop("xmax", None)
        mplkwargs.pop("xstep", None)

        mplkwargs.pop("ymin", None)
        mplkwargs.pop("ymax", None)
        mplkwargs.pop("ystep", None)

        mplkwargs.pop("vmin", None)
        mplkwargs.pop("vmax", None)
        mplkwargs.pop("vstep", None)

        # Remove arguments from mplkwargs that cannot be passed directly into the plotting methods
        plot_args = {"datagroups" : mplkwargs.pop("datagroups", None),
                     "nocolourbar" : mplkwargs.pop("nocolourbar", False),
                     "logx" : mplkwargs.pop("logx", False),
                     "logy" : mplkwargs.pop("logy", False),
                     "logv" : mplkwargs.pop("logv", False),
                     "xrange" : mplkwargs.pop("xrange", {}),
                     "yrange" : mplkwargs.pop("yrange", {}),
                     "valrange" : mplkwargs.pop("valrange", {}),
                     "cbarorient" : mplkwargs.pop("cbarorient", "horizontal"),
                     "grid" : mplkwargs.pop("grid", False),
                     "xlabel" : mplkwargs.pop("xlabel", None),
                     "ylabel" : mplkwargs.pop("ylabel", None),
                     "cbarlabel" : mplkwargs.pop("cbarlabel", None),
                     "title" : mplkwargs.pop("title", None),
                     "fontsize" : mplkwargs.pop("fontsize", None),
                     "itemwidth" : mplkwargs.pop("itemwidth", 1),
                     "xtickangle" : mplkwargs.pop("xtickangle", None),
                     "ytickangle" : mplkwargs.pop("ytickangle", None),
                     "xbinwidth" : mplkwargs.pop("xbinwidth", None),
                     "ybinwidth" : mplkwargs.pop("ybinwidth", None),
                     "coastlinescolour" : mplkwargs.pop("coastlinescolour", "k"),
                     "x_variable" : mplkwargs.pop("x_variable"),
                     "y_variable" : mplkwargs.pop("y_variable")}

        self.mplkwargs = mplkwargs
        self.remove_unassigned_arguments()
        
        if plot_type is None: plot_type = self.set_default_plot_type(packed_data_items)

        # Do plot
        plot = self.plot_types[plot_type](packed_data_items, plot_args, *mplargs, **mplkwargs)
        plot.format_plot()
        plot.apply_axis_limits(plot_args["xrange"], "x")
        plot.apply_axis_limits(plot_args["yrange"], "y")
        self.output_to_file_or_screen(out_filename)

    def output_to_file_or_screen(self, out_filename = None):
        '''
        Outputs to screen unless a filename is given
        @param out_filename: The filename of the file to save the plot to. Various file extensions can be used, with png being the default
        '''
        import logging
        import matplotlib.pyplot as plt
        if out_filename is None:
            plt.show()
        else:
            logging.info("saving plot to file: " + out_filename)
            plt.savefig(out_filename) # Will overwrite if file already exists

    def remove_unassigned_arguments(self):
        '''
        Removes arguments from the mplkwargs if they are equal to None
        '''
        for key in self.mplkwargs.keys():
            if self.mplkwargs[key] is None:
                self.mplkwargs.pop(key)

    def set_default_plot_type(self, data):
        '''
        Sets the default plot type based on the number of dimensions of the data
        @param data: A list of packed data items
        @return The default plot type as a string
        '''
        from jasmin_cis.exceptions import InvalidPlotTypeError
        import logging
        number_of_coords = 0
        for coord in data[0].coords():
            if len(coord.shape) != 1 or coord.shape[0] != 1:
                number_of_coords += 1
        try:
            plot_type = "line" if number_of_coords == 1 else "scatter"
            logging.info("No plot type specified. Plotting data as a " + plot_type)
            return plot_type
        except KeyError:
            coord_shape = None
            all_coords_are_of_same_shape = False
            for coord in data[0].coords():
                if coord_shape == None:
                    coord_shape = coord.shape
                    all_coords_are_of_same_shape = True
                elif coord_shape != coord.shape:
                    all_coords_are_of_same_shape = False
                    break

            error_message = "There is no valid plot type for this variable\nIts shape is: " + str(data[0].shape)
            if all_coords_are_of_same_shape:
                error_message += "\nThe shape of its coordinates is: " + str(data[0].coords()[0].shape)
            raise InvalidPlotTypeError(error_message)


from jasmin_cis.subsetting.subset_framework import SubsetterInterface

class Subsetter(SubsetterInterface):
    def subset(self, data, constraint):
        """Subsets the supplied data using the supplied constraint.

        @param data: data to be subsetted
        @param constraint: SubsetConstraint object to be used to subset data
        @return: subsetted data
        """
        return constraint.constrain(data)



'''
module to test the hdf4 utility function of hdf_sd.py
'''
from nose.tools import istest, eq_
from jasmin_cis.test.test_files.data import *
import jasmin_cis.data_io.hdf_sd as hdf_sd

@istest
def test_that_can_read_all_variables():
    data_dict = hdf_sd.read(valid_hdf_sd_file)
    eq_(len(data_dict),67)

@istest
def test_that_can_read_known_variables():
    data_dict = hdf_sd.read(valid_hdf_sd_file,['Latitude','Longitude'])
    eq_(len(data_dict),2)

@istest
def test_that_can_get_data():
    data_dict = hdf_sd.read(valid_hdf_sd_file)
    data = hdf_sd.get_data(data_dict['Latitude'])
    eq_(data.shape,(203,135))

@istest
def test_that_can_get_metadata_for_known_variable():
    data_dict = hdf_sd.read(valid_hdf_sd_file)
    metadata = hdf_sd.get_metadata(data_dict['Latitude'])

    eq_(metadata._name,"Latitude")
    eq_(metadata.standard_name,"latitude")
    eq_(metadata.long_name,"Geodetic Latitude")
    eq_(metadata.shape,[203, 135])
    eq_(metadata.units,"Degrees_north")
    eq_(metadata.range,[-90.0, 90.0])
    eq_(metadata.factor,1.0)
    eq_(metadata.offset,0.0)
    eq_(metadata.missing_value,-999.0)


    attr = metadata.misc
    eq_(len(attr),10)
    eq_(attr['_FillValue'],-999.0)
    eq_(attr['Parameter_Type'],"MODIS Input")



from nose import with_setup
import os
from jasmin_cis.test.test_util.mock import make_dummy_2d_ungridded_data

tmp_file = "tmp_file"

def delete_tmp_file():
    os.remove(tmp_file)

@with_setup(teardown=delete_tmp_file)
def test_write_netcdf():
    from jasmin_cis.data_io.write_netcdf import write
    data_object = make_dummy_2d_ungridded_data()
    write(data_object, tmp_file)



from nose.tools import istest, raises
from jasmin_cis.info import info
from jasmin_cis.test.test_files.data import *
from jasmin_cis.exceptions import CISError

@istest
def can_get_info_on_a_netcdf_file():
    info(valid_1d_filename)

@istest
def can_get_info_on_a_variable_in_a_netcdf_file():
    info(valid_1d_filename, [valid_variable_in_valid_filename])

@istest
@raises(CISError)
def should_raise_error_when_file_is_not_netcdf():
    info(non_netcdf_file_with_netcdf_file_extension)

@istest
def can_get_info_on_a_cloudsat_file():
    info(valid_cloudsat_RVOD_file)

@istest
def can_get_info_on_a_variable_in_a_cloudsat_file():
    info(valid_cloudsat_RVOD_file, [valid_cloudsat_RVOD_variable])

@istest
def can_get_info_on_a_modis_file():
    info(valid_modis_l2_filename)

@istest
def can_get_info_on_a_variable_in_a_modis_file():
    info(valid_modis_l2_filename, [valid_modis_l2_variable])

@istest
def can_get_info_on_an_aeronet_file():
    info(valid_aeronet_filename)

@istest
def can_get_info_on_a_variable_in_an_aeronet_file():
    info(valid_aeronet_filename, [valid_aeronet_variable])

@istest
def can_get_info_on_a_caliop_file():
    info(valid_caliop_l2_filename)

@istest
def can_get_info_on_a_variable_in_a_caliop_file():
    info(valid_caliop_l2_filename, [valid_caliop_l2_variable])

@istest
def can_get_info_on_a_caliop_l1_file():
    info(valid_caliop_l1_filename)

@istest
def can_get_info_on_a_variable_in_a_caliop_l1_file():
    info(valid_caliop_l1_filename, [valid_caliop_l1_variable])


from jasmin_cis.utils import apply_intersection_mask_to_two_arrays, calculate_histogram_bin_edges
from nose.tools import istest, eq_

@istest
def can_apply_intersection_mask_to_two_masked_arrays():
    import numpy.ma as ma
    import numpy as np
    array1 = ma.array([1,2,3,4,5,6,7,8,9,10], mask=[1,1,1,0,0,0,0,0,0,0])
    array2 = ma.array([2,4,5,6,7,8,4,3,6,80], mask=[0,1,0,0,0,0,0,1,0,0])
    array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
    assert(np.equal(array1.mask, [1,1,1,0,0,0,0,1,0,0]).all())
    assert(np.equal(array2.mask, [1,1,1,0,0,0,0,1,0,0]).all())

@istest
def can_apply_intersection_mask_to_three_masked_arrays():
    import numpy.ma as ma
    array1 = ma.array([1,2,3,4,5,6,7,8,9,10], mask=[1,1,1,0,0,0,0,0,0,0])
    array2 = ma.array([2,4,5,6,7,8,4,3,6,80], mask=[0,1,0,0,0,0,0,1,0,0])
    array3 = ma.array([2,4,5,6,7,8,4,3,6,80], mask=[0,0,0,0,0,0,0,0,0,1])
    array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
    array1, array3 = apply_intersection_mask_to_two_arrays(array1, array3)
    array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)

    assert(ma.equal(array1.mask, [1,1,1,0,0,0,0,1,0,1]).all())
    assert(ma.equal(array2.mask, [1,1,1,0,0,0,0,1,0,1]).all())
    assert(ma.equal(array3.mask, [1,1,1,0,0,0,0,1,0,1]).all())


@istest
def can_apply_intersection_mask_to_one_masked_and_one_unmasked_array():
    import numpy.ma as ma
    import numpy as np
    array1 = ma.array([1,2,3,4,5,6,7,8,9,10], mask=[1,1,1,0,0,0,0,0,0,0])
    array2 = np.array([2,4,5,6,7,8,4,3,6,80])
    array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
    assert(np.equal(array1.mask, [1,1,1,0,0,0,0,0,0,0]).all())
    assert(np.equal(array2.mask,[1,1,1,0,0,0,0,0,0,0]).all())

@istest
def can_apply_intersection_mask_to_two_unmasked_arrays():
    import numpy as np
    array1 = np.array([1,2,3,4,5,6,7,8,9,10])
    array2 = np.array([2,4,5,6,7,8,4,3,6,80])
    array1, array2 = apply_intersection_mask_to_two_arrays(array1, array2)
    assert(all(array1.mask)== False)
    assert(all(array2.mask)== False)

@istest
def can_expand_1d_array_across():
    import numpy as np
    from jasmin_cis.utils import expand_1d_to_2d_array
    a = np.array([1,2,3,4])
    b = expand_1d_to_2d_array(a, 4, axis=0)
    ref = np.array([[1, 2, 3, 4],[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
    assert(np.equal(b, ref).all())

@istest
def can_expand_1d_array_down():
    import numpy as np
    from jasmin_cis.utils import expand_1d_to_2d_array
    a = np.array([1,2,3,4])
    b = expand_1d_to_2d_array(a, 4, axis=1)
    ref = np.array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]])
    assert(np.equal(b, ref).all())

@istest
def ten_bins_are_created_by_default():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {}
    step = None

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 11) # 11 edges = 10 bins
    eq_(bin_edges.min(), data.min())
    assert(abs(bin_edges.max() - data.max()) < 1.e-7)

@istest
def bin_width_can_be_specified_where_bin_width_perfectly_divides_range():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {}
    step = 0.5

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 7)
    eq_(bin_edges.min(), data.min())
    eq_(bin_edges.max(), data.max())

@istest
def bin_width_can_be_specified_where_bin_width_does_not_perfectly_divides_range():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {}
    step = 0.7

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 5)
    eq_(bin_edges.min(), data.min())
    assert(bin_edges.max() < data.max())

@istest
def ten_bins_are_created_when_min_is_specified():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {"xmin" : 0.3}
    step = None

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 11) # 11 edges = 10 bins
    eq_(bin_edges.min(), 0.3)
    assert(abs(bin_edges.max() - data.max()) < 1.e-7) # 1.e-7 is approx 0

@istest
def ten_bins_are_created_when_max_is_specified():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {"xmax" : 2.3}
    step = None

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 11) # 11 edges = 10 bins
    eq_(bin_edges.min(), data.min())
    assert(abs(bin_edges.max() - 2.3) < 1.e-7) # 1.e-7 is approx 0'''

@istest
def ten_bins_are_created_when_min_and_max_is_specified():
    from numpy import array
    data = array([0.0, 1.0, 2.0, 3.0])
    val_range = {"xmin" : 0.3, "xmax" : 2.3}
    step = None

    bin_edges = calculate_histogram_bin_edges(data, "x", val_range, step)
    eq_(len(bin_edges), 11) # 11 edges = 10 bins
    assert(abs(bin_edges.min() - 0.3) < 1.e-7) # 1.e-7 is approx 0
    assert(abs(bin_edges.max() - 2.3) < 1.e-7) # 1.e-7 is approx 0
